# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import numpy as np


def prepare_for_fortran(overwrite, *args):
    """Convert arrays to Fortran format.

    This function takes a number of array objects in `args` and converts them
    to a format that can be directly passed to a Fortran function (Fortran
    contiguous NumPy array). If the arrays have different data type, they
    converted arrays are cast to a common compatible data type (one of NumPy's
    `float32`, `float64`, `complex64`, `complex128` data types).

    If `overwrite` is ``False``, an NumPy array that would already be in the
    correct format (Fortran contiguous, right data type) is neverthelessed
    copied. (Hence, overwrite = True does not imply that acting on the
    converted array in the return values will overwrite the original array in
    all cases -- it does only so if the original array was already in the
    correct format. The conversions require copying. In fact, that's the same
    behavior as in SciPy, it's just not explicitly stated there)

    If an argument is ``None``, it is just passed through and not used to
    determine the proper data type.

    `prepare_for_lapack` returns a character indicating the proper
    data type in LAPACK style ('s', 'd', 'c', 'z') and a list of
    properly converted arrays.
    """

    # Make sure we have NumPy arrays
    mats = [None]*len(args)
    for i in range(len(args)):
        if args[i] is not None:
            arr = np.asanyarray(args[i])
            if not np.issubdtype(arr.dtype, np.number):
                raise ValueError("Argument cannot be interpreted "
                                 "as a numeric array")

            mats[i] = (arr, arr is not args[i] or overwrite)
        else:
            mats[i] = (None, True)

    # First figure out common dtype
    # Note: The return type of common_type is guaranteed to be a floating point
    #       kind.
    dtype = np.common_type(*[arr for arr, ovwrt in mats if arr is not None])

    if dtype == np.float32:
        lapacktype = 's'
    elif dtype == np.float64:
        lapacktype = 'd'
    elif dtype == np.complex64:
        lapacktype = 'c'
    elif dtype == np.complex128:
        lapacktype = 'z'
    else:
        raise AssertionError("Unexpected data type from common_type")

    ret = [ lapacktype ]
    for npmat, ovwrt in mats:
        # Now make sure that the array is contiguous, and copy if necessary.
        if npmat is not None:
            if npmat.ndim == 2:
                if not npmat.flags["F_CONTIGUOUS"]:
                    npmat = np.asfortranarray(npmat, dtype = dtype)
                elif npmat.dtype != dtype:
                    npmat = npmat.astype(dtype)
                elif not ovwrt:
                    # ugly here: copy makes always C-array, no way to tell it
                    # to make a Fortran array.
                    npmat = np.asfortranarray(npmat.copy())
            elif npmat.ndim == 1:
                if not npmat.flags["C_CONTIGUOUS"]:
                    npmat = np.ascontiguousarray(npmat, dtype = dtype)
                elif npmat.dtype != dtype:
                    npmat = npmat.astype(dtype)
                elif not ovwrt:
                    npmat = np.asfortranarray(npmat.copy())
            else:
                raise ValueError("Dimensionality of array is not 1 or 2")

        ret.append(npmat)

    return tuple(ret)


def assert_fortran_mat(*mats):
    """Check if the input ndarrays are all proper Fortran matrices."""

    # This is a workaround for a bug in NumPy version < 2.0,
    # where 1x1 matrices do not have the F_Contiguous flag set correctly.
    for mat in mats:
        if (mat is not None and (mat.shape[0] > 1 or mat.shape[1] > 1) and
            not mat.flags["F_CONTIGUOUS"]):
            raise ValueError("Input matrix must be Fortran contiguous")


def assert_fortran_matvec(*arrays):
    """Check if the input ndarrays are all proper Fortran matrices
    or vectors."""

    # This is a workaround for a bug in NumPy version < 2.0,
    # where 1x1 matrices do not have the F_Contiguous flag set correctly.
    for arr in arrays:
        if not arr.ndim in (1, 2):
            raise ValueError("Input must be either a vector "
                             "or a matrix.")

        if (not arr.flags["F_CONTIGUOUS"] or
            (arr.ndim == 2 and arr.shape[0] == 1 and arr.shape[1] == 1) ):
            raise ValueError("Input must be a Fortran ordered "
                             "NumPy array")



import math
from cmath import exp
import kwant


def hopping(sitei, sitej, phi, salt):
    xi, yi = sitei.pos
    xj, yj = sitej.pos
    return -exp(-0.5j * phi * (xi - xj) * (yi + yj))


def onsite(site, phi, salt):
    return 0.3 * kwant.digest.gauss(repr(site), salt) + 4


def test_qhe(W=16, L=8):
    def central_region(pos):
        x, y = pos
        return -L < x < L and abs(y) < W - 5.5 * math.exp(-x**2 / 5**2)

    lat = kwant.lattice.square()
    syst = kwant.Builder()

    syst[lat.shape(central_region, (0, 0))] = onsite
    syst[lat.neighbors()] = hopping

    lead = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    lead[(lat(0, y) for y in range(-W + 1, W))] = 4
    lead[lat.neighbors()] = hopping

    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())
    syst = syst.finalized()

    #### The following chunk of code can be uncommented to visualize the
    #### conductance plateaus.
    # from matplotlib import pyplot
    # import numpy
    # reciprocal_phis = numpy.linspace(0.1, 7, 200)
    # conductances = []
    # for phi in 1 / reciprocal_phis:
    #     smatrix = kwant.smatrix(syst, 1.0, [phi, ""])
    #     conductances.append(smatrix.transmission(1, 0))
    # pyplot.plot(reciprocal_phis, conductances)
    # pyplot.show()

    for r_phis, T_nominal, max_err in [((1.3, 2.1), 1, 1e-7),
                                       ((3.2, 3.7), 2, 1e-3),
                                       ((5.2, 5.5), 3, 1e-1)]:
        for r_phi in r_phis:
            params = dict(phi=1.0 / r_phi, salt="")
            pc = syst.precalculate(1.0, params=params, what='all')
            for result in [kwant.smatrix(pc, 1, params=params),
                           kwant.solvers.default.greens_function(pc, 1, params=params)]:
                assert abs(T_nominal - result.transmission(1, 0)) < max_err



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

__all__ = ['lu_factor', 'lu_solve', 'rcond_from_lu']

import numpy as np
from . import lapack


def lu_factor(a, overwrite_a=False):
    """Compute the LU factorization of a matrix A = P * L * U. The function
    returns a tuple (lu, p, singular), where lu contains the LU factorization
    storing the unit lower triangular matrix L in the strictly lower triangle
    (the unit diagonal is not stored) and the upper triangular matrix U in the
    upper triangle. p is a vector of pivot indices, and singular a Boolean
    value indicating whether the matrix A is singular up to machine precision.

    NOTE: This function mimics the behavior of scipy.linalg.lu_factor (except
    that it has in addition the flag singular). The main reason is that
    lu_factor in SciPy has a bug that depending on the type of NumPy matrix
    passed to it, it would not return what was descirbed in the
    documentation. This bug will be (probably) fixed in 0.10.0 but until this
    is standard, this version is better to use.

    Parameters
    ----------
    a : array, shape (M, M)
        Matrix to factorize
    overwrite_a : boolean
        Whether to overwrite data in a (may increase performance)

    Returns
    -------
    lu : array, shape (N, N)
        Matrix containing U in its upper triangle, and L in its lower triangle.
        The unit diagonal elements of L are not stored.
    piv : array, shape (N,)
        Pivot indices representing the permutation matrix P:
        row i of matrix was interchanged with row piv[i].
    singular : boolean
        Whether the matrix a is singular (up to machine precision)
    """
    a = lapack.prepare_for_lapack(overwrite_a, a)
    return lapack.getrf(a)


def lu_solve(matrix_factorization, b):
    """Solve a linear system of equations, a x = b, given the LU
    factorization of a

    Parameters
    ----------
    matrix_factorization
        Factorization of the coefficient matrix a, as given by lu_factor
    b : array (vector or matrix)
        Right-hand side

    Returns
    -------
    x : array (vector or matrix)
        Solution to the system
    """
    (lu, ipiv, singular) = matrix_factorization
    if singular:
        raise RuntimeWarning("In lu_solve: the flag singular indicates "
                             "a singular matrix. Result of solve step "
                             "are probably unreliable")

    lu, b = lapack.prepare_for_lapack(False, lu, b)
    ipiv = np.ascontiguousarray(np.asanyarray(ipiv), dtype=lapack.int_dtype)
    return lapack.getrs(lu, ipiv, b)


def rcond_from_lu(matrix_factorization, norm_a, norm="1"):
    """Compute the reciprocal condition number from the LU decomposition as
    returned from lu_factor(), given additionally the norm of the matrix a in
    norm_a.

    The reciprocal condition number is given as 1/(||A||*||A^-1||), where
    ||...|| is a matrix norm.

    Parameters
    ----------
    matrix_factorization
        Factorization of the matrix a, as given by lu_factor
    norm_a : float or complex
        norm of the original matrix a (type of norm is specified in norm)
    norm : {'1', 'I'}, optional
        type of matrix norm which should be used to compute the condition
        number ("1": 1-norm, "I": infinity norm). Default: '1'.

    Returns
    -------
    rcond : float or complex
        reciprocal condition number of a with respect to the type of matrix
        norm specified in norm
    """
    (lu, ipiv, singular) = matrix_factorization
    norm = norm.encode('utf8')  # lapack expects bytes
    lu = lapack.prepare_for_lapack(False, lu)
    return lapack.gecon(lu, norm_a, norm)



# Copyright 2011-2018 Kwant authors.
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

"""Physics-related algorithms"""

# Merge the public interface of all submodules.
__all__ = []
for module in ['leads', 'dispersion', 'noise', 'symmetry', 'gauge']:
    exec('from . import {0}'.format(module))
    exec('from .{0} import *'.format(module))
    exec('__all__.extend({0}.__all__)'.format(module))



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

"""Interface to the MUMPS sparse solver library"""

__all__ = ['MUMPSContext', 'schur_complement', 'AnalysisStatistics',
           'FactorizationStatistics', 'MUMPSError']

import time
import numpy as np
import scipy.sparse
import warnings
from . import _mumps
from .fortran_helpers import prepare_for_fortran

orderings = { 'amd' : 0, 'amf' : 2, 'scotch' : 3, 'pord' : 4, 'metis' : 5,
              'qamd' : 6, 'auto' : 7 }

ordering_name = [ 'amd', 'user-defined', 'amf',
                  'scotch', 'pord', 'metis', 'qamd']


def possible_orderings():
    """Return the ordering options that are available in the current
    installation of MUMPS.

    Which ordering options are actually available depends how MUMPs was
    compiled. Note that passing an ordering that is not avaialble in the
    current installation of MUMPS will not fail, instead MUMPS will fall back
    to a supported one.

    Returns
    -------
    orderings : list of strings
       A list of installed orderings that can be used in the `ordering` option
       of MUMPS.
    """

    if not possible_orderings.cached:
        # Try all orderings on a small test matrix, and check which one was
        # actually used.

        possible_orderings.cached = ['auto']
        for ordering in [0, 2, 3, 4, 5, 6]:
            data = np.asfortranarray([1, 1], dtype=np.complex128)
            row = np.asfortranarray([1, 2], dtype=_mumps.int_dtype)
            col = np.asfortranarray([1, 2], dtype=_mumps.int_dtype)

            instance = _mumps.zmumps()
            instance.set_assembled_matrix(2, row, col, data)
            instance.icntl[7] = ordering
            instance.job = 1
            instance.call()

            if instance.infog[7] == ordering:
                possible_orderings.cached.append(ordering_name[ordering])

    return possible_orderings.cached

possible_orderings.cached = None


error_messages = {
    -5 : "Not enough memory during analysis phase",
    -6 : "Matrix is singular in structure",
    -7 : "Not enough memory during analysis phase",
    -10 : "Matrix is numerically singular",
    -11 : "The authors of MUMPS would like to hear about this",
    -12 : "The authors of MUMPS would like to hear about this",
    -13 : "Not enough memory"
}

class MUMPSError(RuntimeError):
    def __init__(self, infog):
        self.error = infog[1]
        if self.error in error_messages:
            msg = "{}. (MUMPS error {})".format(
                error_messages[self.error], self.error)
        else:
            msg = "MUMPS failed with error {}.".format(self.error)

        RuntimeError.__init__(self, msg)


class AnalysisStatistics:
    def __init__(self, inst, time=None):
        self.est_mem_incore = inst.infog[17]
        self.est_mem_ooc = inst.infog[27]
        self.est_nonzeros = (inst.infog[20] if inst.infog[20] > 0 else
                             -inst.infog[20] * 1000000)
        self.est_flops = inst.rinfog[1]
        self.ordering = ordering_name[inst.infog[7]]
        self.time = time

    def __str__(self):
        parts = ["estimated memory for in-core factorization:",
                 str(self.est_mem_incore), "mbytes\n",
                 "estimated memory for out-of-core factorization:",
                 str(self.est_mem_ooc), "mbytes\n",
                 "estimated number of nonzeros in factors:",
                 str(self.est_nonzeros), "\n",
                 "estimated number of flops:", str(self.est_flops), "\n",
                 "ordering used:", self.ordering]
        if hasattr(self, "time"):
            parts.extend(["\n analysis time:", str(self.time), "secs"])
        return " ".join(parts)


class FactorizationStatistics:
    def __init__(self, inst, time=None, include_ordering=False):
        # information about pivoting
        self.offdiag_pivots = inst.infog[12] if inst.sym == 0 else 0
        self.delayed_pivots = inst.infog[13]
        self.tiny_pivots = inst.infog[25]

        # possibly include ordering (used in schur_complement)
        if include_ordering:
            self.ordering = ordering_name[inst.infog[7]]

        # information about runtime effiency
        self.memory = inst.infog[22]
        self.nonzeros = (inst.infog[29] if inst.infog[29] > 0 else
                         -inst.infog[29] * 1000000)
        self.flops = inst.rinfog[3]
        if time:
            self.time = time

    def __str__(self):
        parts = ["off-diagonal pivots:", str(self.offdiag_pivots), "\n",
                 "delayed pivots:", str(self.delayed_pivots), "\n",
                 "tiny pivots:", str(self.tiny_pivots), "\n"]
        if hasattr(self, "ordering"):
            parts.extend(["ordering used:", self.ordering, "\n"])
        parts.extend(["memory used during factorization:", str(self.memory),
                      "mbytes\n",
                      "nonzeros in factored matrix:", str(self.nonzeros), "\n",
                      "floating point operations:", str(self.flops)])
        if hasattr(self, "time"):
            parts.extend(["\n factorization time:", str(self.time), "secs"])
        return " ".join(parts)


class MUMPSContext:
    """MUMPSContext contains the internal data structures needed by the
    MUMPS library and contains a user-friendly interface.

    WARNING: Only complex numbers supported.

    Examples
    --------

    Solving a small system of equations.

    >>> import scipy.sparse as sp
    >>> a = sp.coo_matrix([[1.,0],[0,2.]], dtype=complex)
    >>> ctx = kwant.linalg.mumps.MUMPSContext()
    >>> ctx.factor(a)
    >>> ctx.solve([1., 1.])
    array([ 1.0+0.j,  0.5+0.j])

    Instance variables
    ------------------

    analysis_stats : `AnalysisStatistics`
        contains MUMPS statistics after an analysis step (i.e.  after a call to
        `analyze` or `factor`)
    factor_stats : `FactorizationStatistics`
        contains MUMPS statistics after a factorization step (i.e.  after a
        call to `factor`)

    """

    def __init__(self, verbose=False):
        """Init the MUMPSContext class

        Parameters
        ----------

        verbose : True or False
            control whether MUMPS prints lots of internal statistics
            and debug information to screen.
        """
        self.mumps_instance = None
        self.dtype = None
        self.verbose = verbose
        self.factored = False

    def analyze(self, a, ordering='auto', overwrite_a=False):
        """Perform analysis step of MUMPS.

        In the analyis step, MUMPS figures out a reordering for the matrix and
        estimates number of operations and memory needed for the factorization
        time. This step usually needs not be called separately (it is done
        automatically by `factor`), but it can be useful to test which ordering
        would give best performance in the actual factorization, as MUMPS
        estimates are available in `analysis_stats`.

        Parameters
        ----------

        a : sparse SciPy matrix
            input matrix. Internally, the matrix is converted to `coo` format
            (so passing this format is best for performance)
        ordering : { 'auto', 'amd', 'amf', 'scotch', 'pord', 'metis', 'qamd' }
            ordering to use in the factorization. The availability of a
            particular ordering depends on the MUMPS installation.  Default is
            'auto'.
        overwrite_a : True or False
            whether the data in a may be overwritten, which can lead to a small
            performance gain. Default is False.
        """

        a = a.tocoo()

        if a.ndim != 2 or a.shape[0] != a.shape[1]:
            raise ValueError("Input matrix must be square!")

        if not ordering in orderings.keys():
            raise ValueError("Unknown ordering '"+ordering+"'!")

        dtype, row, col, data = _make_assembled_from_coo(a, overwrite_a)

        if dtype != self.dtype:
            self.mumps_instance = getattr(_mumps, dtype+"mumps")(self.verbose)
            self.dtype = dtype

        self.n = a.shape[0]
        self.row = row
        self.col = col
        self.data = data
        # Note: if I don't store them, they go out of scope and are
        #       deleted. I however need the memory to stay around!

        self.mumps_instance.set_assembled_matrix(a.shape[0], row, col, data)
        self.mumps_instance.icntl[7] = orderings[ordering]
        self.mumps_instance.job = 1
        t1 = time.process_time()
        self.mumps_instance.call()
        t2 = time.process_time()
        self.factored = False

        if self.mumps_instance.infog[1] < 0:
            raise MUMPSError(self.mumps_instance.infog)

        self.analysis_stats = AnalysisStatistics(self.mumps_instance,
                                                 t2 - t1)

    def factor(self, a, ordering='auto', ooc=False, pivot_tol=0.01,
               reuse_analysis=False, overwrite_a=False):
        """Perform the LU factorization of the matrix.

        This LU factorization can then later be used to solve a linear system
        with `solve`. Statistical data of the factorization is stored in
        `factor_stats`.

        Parameters
        ----------

        a : sparse SciPy matrix
            input matrix. Internally, the matrix is converted to `coo` format
            (so passing this format is best for performance)
        ordering : { 'auto', 'amd', 'amf', 'scotch', 'pord', 'metis', 'qamd' }
            ordering to use in the factorization. The availability of a
            particular ordering depends on the MUMPS installation.  Default is
            'auto'.
        ooc : True or False
            whether to use the out-of-core functionality of MUMPS.
            (out-of-core means that data is written to disk to reduce memory
            usage.) Default is False.
        pivot_tol: number in the range [0, 1]
            pivoting threshold. Pivoting is typically limited in sparse
            solvers, as too much pivoting destroys sparsity. 1.0 means full
            pivoting, whereas 0.0 means no pivoting. Default is 0.01.
        reuse_analysis: True or False
            whether to reuse the analysis done in a previous call to `analyze`
            or `factor`. If the structure of the matrix stays the same, and the
            numerical values do not change much, the previous analysis can be
            reused, saving some time.  WARNING: There is no check whether the
            structure of your matrix is compatible with the previous
            analysis. Also, if the values are not similar enough, there might
            be loss of accuracy, without a warning. Default is False.
        overwrite_a : True or False
            whether the data in a may be overwritten, which can lead to a small
            performance gain. Default is False.
        """
        a = a.tocoo()

        if a.ndim != 2 or a.shape[0] != a.shape[1]:
            raise ValueError("Input matrix must be square!")

        # Analysis phase must be done before factorization
        # Note: previous analysis is reused only if reuse_analysis == True

        if reuse_analysis:
            if self.mumps_instance is None:
                warnings.warn("Missing analysis although reuse_analysis=True. "
                              "New analysis is performed.",
                              RuntimeWarning,
                              stacklevel=2)
                self.analyze(a, ordering=ordering, overwrite_a=overwrite_a)
            else:
                dtype, row, col, data = _make_assembled_from_coo(a,
                                                                 overwrite_a)
                if self.dtype != dtype:
                    raise ValueError("MUMPSContext dtype and matrix dtype "
                                     "incompatible!")

                self.n = a.shape[0]
                self.row = row
                self.col = col
                self.data = data
                self.mumps_instance.set_assembled_matrix(a.shape[0],
                                                         row, col, data)
        else:
            self.analyze(a, ordering=ordering, overwrite_a=overwrite_a)

        self.mumps_instance.icntl[22] = 1 if ooc else 0
        self.mumps_instance.job = 2
        self.mumps_instance.cntl[1] = pivot_tol

        done = False
        while not done:
            t1 = time.process_time()
            self.mumps_instance.call()
            t2 = time.process_time()

            # error -8, -9 (not enough allocated memory) is treated
            # specially, by increasing the memory relaxation parameter
            if self.mumps_instance.infog[1] < 0:
                if self.mumps_instance.infog[1] in (-8, -9):
                    # double the additional memory
                    self.mumps_instance.icntl[14] *= 2
                else:
                    raise MUMPSError(self.mumps_instance.infog)
            else:
                done = True

        self.factored = True
        self.factor_stats = FactorizationStatistics(self.mumps_instance,
                                                    t2 - t1)

    def _solve_sparse(self, b):
        b = b.tocsc()
        x = np.empty((b.shape[0], b.shape[1]),
                     order='F', dtype=self.data.dtype)

        dtype, col_ptr, row_ind, data = _make_sparse_rhs_from_csc(
            b, self.data.dtype)

        if b.shape[0] != self.n:
            raise ValueError("Right hand side has wrong size")

        if self.dtype != dtype:
            raise ValueError("Data type of right hand side is not "
                             "compatible with the dtype of the "
                             "linear system")

        self.mumps_instance.set_sparse_rhs(col_ptr, row_ind, data)
        self.mumps_instance.set_dense_rhs(x)
        self.mumps_instance.job = 3
        self.mumps_instance.icntl[20] = 1
        self.mumps_instance.call()

        return x

    def _solve_dense(self, b, overwrite_b=False):
        dtype, b = prepare_for_fortran(overwrite_b, b,
                                       np.zeros(1, dtype=self.data.dtype))[:2]

        if b.shape[0] != self.n:
            raise ValueError("Right hand side has wrong size")

        if self.dtype != dtype:
            raise ValueError("Data type of right hand side is not "
                             "compatible with the dtype of the "
                             "linear system")

        self.mumps_instance.set_dense_rhs(b)
        self.mumps_instance.job = 3
        self.mumps_instance.call()

        return b

    def solve(self, b, overwrite_b=False):
        """Solve a linear system after the LU factorization has previously
        been performed by `factor`.

        Supports both dense and sparse right hand sides.

        Parameters
        ----------

        b : dense (NumPy) matrix or vector or sparse (SciPy) matrix
            the right hand side to solve. Accepts both dense and sparse input;
            if the input is sparse 'csc' format is used internally (so passing
            a 'csc' matrix gives best performance).
        overwrite_b : True or False
            whether the data in b may be overwritten, which can lead to a small
            performance gain. Default is False.

        Returns
        -------

        x : NumPy array
            the solution to the linear system as a dense matrix (a vector is
            returned if b was a vector, otherwise a matrix is returned).
        """

        if not self.factored:
            raise RuntimeError("Factorization must be done before solving!")

        if scipy.sparse.isspmatrix(b):
            return self._solve_sparse(b)
        else:
            return self._solve_dense(b, overwrite_b)


def schur_complement(a, indices, ordering='auto', ooc=False, pivot_tol=0.01,
                     calc_stats=False, overwrite_a=False):
    """Compute the Schur complement block of matrix a using MUMPS.

    Parameters:
    a : sparse matrix
        input matrix. Internally, the matrix is converted to `coo` format (so
        passing this format is best for performance)
    indices : 1d array
        indices (row and column) of the desired Schur complement block.  (The
        Schur complement block is square, so that the indices are both row and
        column indices.)
    ordering : { 'auto', 'amd', 'amf', 'scotch', 'pord', 'metis', 'qamd' }
        ordering to use in the factorization. The availability of a particular
        ordering depends on the MUMPS installation.  Default is 'auto'.
    ooc : True or False
        whether to use the out-of-core functionality of MUMPS.  (out-of-core
        means that data is written to disk to reduce memory usage.) Default is
        False.
    pivot_tol: number in the range [0, 1]
        pivoting threshold. Pivoting is typically limited in sparse solvers, as
        too much pivoting destroys sparsity. 1.0 means full pivoting, whereas
        0.0 means no pivoting. Default is 0.01.
    calc_stats: True or False
        whether to return the analysis and factorization statistics collected
        by MUMPS. Default is False.
    overwrite_a : True or False
        whether the data in a may be overwritten, which can lead to a small
        performance gain. Default is False.

    Returns
    -------

    s : NumPy array
        Schur complement block
    factor_stats: `FactorizationStatistics`
        statistics of the factorization as collected by MUMPS.  Only returned
        if ``calc_stats==True``.
    """

    if not scipy.sparse.isspmatrix(a):
        raise ValueError("a must be a sparse SciPy matrix!")

    a = a.tocoo()

    if a.ndim != 2 or a.shape[0] != a.shape[1]:
        raise ValueError("Input matrix must be square!")

    indices = np.asanyarray(indices)

    if indices.ndim != 1:
        raise ValueError("Schur indices must be specified in a 1d array!")

    if not ordering in orderings.keys():
        raise ValueError("Unknown ordering '"+ordering+"'!")

    dtype, row, col, data = _make_assembled_from_coo(a, overwrite_a)
    indices = _make_mumps_index_array(indices)

    mumps_instance = getattr(_mumps, dtype+"mumps")()

    mumps_instance.set_assembled_matrix(a.shape[0], row, col, data)
    mumps_instance.icntl[7] = orderings[ordering]
    mumps_instance.icntl[19] = 1
    mumps_instance.icntl[31] = 1  # discard factors, from 4.10.0
                                  # has no effect in earlier versions

    schur_compl = np.empty((indices.size, indices.size),
                           order='C', dtype=data.dtype)
    mumps_instance.set_schur(schur_compl, indices)

    mumps_instance.job = 4   # job=4 -> 1 and 2 after each other
    t1 = time.process_time()
    mumps_instance.call()
    t2 = time.process_time()

    if not calc_stats:
        return schur_compl
    else:
        return (schur_compl, FactorizationStatistics(
            mumps_instance, time=t2 - t1, include_ordering=True))


# Some internal helper functions
def _make_assembled_from_coo(a, overwrite_a):
    dtype, data = prepare_for_fortran(overwrite_a, a.data)

    row = np.asfortranarray(a.row.astype(_mumps.int_dtype))
    col = np.asfortranarray(a.col.astype(_mumps.int_dtype))

    # MUMPS uses Fortran indices.
    row += 1
    col += 1

    return dtype, row, col, data


def _make_sparse_rhs_from_csc(b, dtype):
    dtype, data = prepare_for_fortran(True, b.data,
                                      np.zeros(1, dtype=dtype))[:2]

    col_ptr = np.asfortranarray(b.indptr.astype(_mumps.int_dtype))
    row_ind = np.asfortranarray(b.indices.astype(_mumps.int_dtype))

    # MUMPS uses Fortran indices.
    col_ptr += 1
    row_ind += 1

    return dtype, col_ptr, row_ind, data


def _make_mumps_index_array(a):
    a = np.asfortranarray(a.astype(_mumps.int_dtype))
    a += 1                      # Fortran indices

    return a



# Copyright 2011-2017 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

from keyword import iskeyword
from collections import defaultdict
import itertools
import warnings

import numpy as np
import tinyarray as ta

import sympy
from sympy.matrices.matrices import MatrixBase
from sympy.utilities.lambdify import lambdastr
from sympy.printing.lambdarepr import LambdaPrinter
from sympy.printing.precedence import precedence
from sympy.core.function import AppliedUndef

from .. import builder, lattice
from .. import KwantDeprecationWarning
from .._common import reraise_warnings
from ._common import (sympify, gcd, position_operators, momentum_operators,
                      monomials)


__all__ = ['discretize']


_wf = sympy.Function('_internal_unique_name', commutative=False)
_momentum_operators = {s.name: s for s in momentum_operators}
_position_operators = {s.name: s for s in position_operators}
_displacements = {s: sympy.Symbol('a_{}'.format(s)) for s in 'xyz'}


class _DiscretizedBuilder(builder.Builder):
    """A builder that is made from a discretized model and knows how to
    pretty-print itself."""

    def __init__(self, coords, lattice, symmetry=None, **kwargs):
        super().__init__(symmetry, **kwargs)
        self.lattice = lattice
        self._coords = coords

    def __str__(self):
        result = []

        sv = list(s for s in self.site_value_pairs())
        if len(sv) != 1:
            raise ValueError("Cannot pretty-print _DiscretizedBuilder: "
                             "must contain a single site.")
        site, site_value = sv[0]
        if any(e != 0 for e in site.tag):
            raise ValueError("Cannot pretty-print _DiscretizedBuilder: "
                             "site must be located at origin.")

        result.extend(["# Discrete coordinates: ",
                       " ".join(self._coords),
                       "\n\n"])

        for key, val in itertools.chain(self.site_value_pairs(),
                                        self.hopping_value_pairs()):
            if isinstance(key, builder.Site):
                result.append("# Onsite element:\n")
            else:
                a, b = key
                assert a is site
                result.extend(["# Hopping from ",
                               str(tuple(b.tag)),
                               ":\n"])
            result.append(val._source if callable(val) else repr(val))
            result.append('\n\n')

        result.pop()

        return "".join(result)

    # Enable human-readable rendering in Jupyter notebooks:
    def _repr_html_(self):
        return self.__str__()


################ Interface functions


def discretize(hamiltonian, coords=None, *, grid=None, locals=None,
               grid_spacing=None):
    """Construct a tight-binding model from a continuum Hamiltonian.

    If necessary, the given Hamiltonian is sympified using
    `kwant.continuum.sympify`.  It is then discretized symbolically and turned
    into a `~kwant.builder.Builder` instance that may be used with
    `~kwant.builder.Builder.fill`.

    This is a convenience function that is equivalent to first calling
    `~kwant.continuum.discretize_symbolic` and feeding its result into
    `~kwant.continuum.build_discretized`.

    .. warning::
        This function uses ``eval`` (because it calls ``sympy.sympify``), and
        thus should not be used on unsanitized input.

    Parameters
    ----------
    hamiltonian : str or SymPy expression
        Symbolic representation of a continuous Hamiltonian.  It is
        converted to a SymPy expression using `kwant.continuum.sympify`.
    coords : sequence of strings, optional
        The coordinates for which momentum operators will be treated as
        differential operators. May contain only "x", "y" and "z" and must be
        sorted.  If not provided, `coords` will be obtained from the input
        Hamiltonian by reading the present coordinates and momentum operators.
    grid : scalar or kwant.lattice.Monatomic instance, optional
        Lattice that will be used as a discretization grid. It must have
        orthogonal primitive vectors. If a scalar value is given, a lattice
        with the appropriate grid spacing will be generated. If not provided,
        a lattice with grid spacing 1 in all directions will be generated.
    locals : dict, optional
        Additional namespace entries for `~kwant.continuum.sympify`.  May be
        used to simplify input of matrices or modify input before proceeding
        further. For example:
        ``locals={'k': 'k_x + I * k_y'}`` or
        ``locals={'sigma_plus': [[0, 2], [0, 0]]}``.
    grid_spacing : int or float, optional
        (deprecated) Spacing of the discretization grid. If unset the default
        value will be 1. Cannot be used together with ``grid``.

    Returns
    -------
    model : `~kwant.builder.Builder`
        The translationally symmetric builder that corresponds to the provided
        Hamiltonian.  This builder instance belongs to a subclass of the
        standard builder that may be printed to obtain the source code of the
        value functions.  It also holds the discretization lattice (a
        `~kwant.lattice.Monatomic` instance with lattice constant
        `grid_spacing`) in the ``lattice`` attribute.
    """
    tb, coords = discretize_symbolic(hamiltonian, coords, locals=locals)
    return build_discretized(tb, coords, grid=grid, grid_spacing=grid_spacing)


def discretize_symbolic(hamiltonian, coords=None, *, locals=None):
    """Discretize a continuous Hamiltonian into a tight-binding representation.

    If necessary, the given Hamiltonian is sympified using
    `kwant.continuum.sympify`.  It is then discretized symbolically.

    The two return values may be used directly as the first two arguments for
    `~kwant.continuum.build_discretized`.

    .. warning::
        This function uses ``eval`` (because it calls ``sympy.sympify``), and
        thus should not be used on unsanitized input.

    Parameters
    ----------
    hamiltonian : str or SymPy expression
        Symbolic representation of a continuous Hamiltonian.  It is
        converted to a SymPy expression using `kwant.continuum.sympify`.
    coords : sequence of strings, optional
        The coordinates for which momentum operators will be treated as
        differential operators. May contain only "x", "y" and "z" and must be
        sorted.  If not provided, `coords` will be obtained from the input
        Hamiltonian by reading the present coordinates and momentum operators.
    locals : dict, optional
        Additional namespace entries for `~kwant.continuum.sympify`.  May be
        used to simplify input of matrices or modify input before proceeding
        further. For example:
        ``locals={'k': 'k_x + I * k_y'}`` or
        ``locals={'sigma_plus': [[0, 2], [0, 0]]}``.

    Returns
    -------
    tb_hamiltonian : dict
        Keys are tuples of integers; the offsets of the hoppings ((0, 0, 0) for
        the onsite). Values are symbolic expressions for the hoppings/onsite.
    coords : list of strings
        The coordinates that have been discretized.
    """
    with reraise_warnings():
        hamiltonian = sympify(hamiltonian, locals)

    atoms_names = [s.name for s in hamiltonian.atoms(sympy.Symbol)]
    if any(s in ('a_x', 'a_y', 'a_z') for s in atoms_names):
        raise TypeError("'a_x', 'a_y' and 'a_z' are  symbols used internally "
                        "to represent grid spacings; please use a different "
                        "symbol.")

    hamiltonian = sympy.expand(hamiltonian)
    if coords is None:
        used_momenta = set(_momentum_operators) & set(atoms_names)
        coords = {k[-1] for k in used_momenta}
    else:
        coords = list(coords)
        if coords != sorted(coords):
            raise ValueError("The argument 'coords' must be sorted.")
        if any(c not in 'xyz' for c in coords):
            raise ValueError("The argument 'coords' may only contain "
                             "'x', 'y', or 'z'.")

    coords = sorted(coords)

    if len(coords) == 0:
        raise ValueError("Failed to read any discrete coordinates. This is "
                         "probably due to a lack of momentum operators in "
                         "your input. You can use the 'coords' "
                         "parameter to provide them.")

    onsite_zeros = (0,) * len(coords)

    if not isinstance(hamiltonian, MatrixBase):
        hamiltonian = sympy.Matrix([hamiltonian])
        _input_format = 'expression'
    else:
        _input_format = 'matrix'

    shape = hamiltonian.shape
    tb = defaultdict(lambda: sympy.zeros(*shape))
    tb[onsite_zeros] = sympy.zeros(*shape)

    for (i, j), expression in np.ndenumerate(hamiltonian):
        hoppings = _discretize_expression(expression, coords)

        for offset, hop in hoppings.items():
            tb[offset][i, j] += hop

    # do not include Hermitian conjugates of hoppings
    wanted_hoppings = sorted(list(tb))[len(list(tb)) // 2:]
    tb = {k: v for k, v in tb.items() if k in wanted_hoppings}

    if _input_format == 'expression':
        tb = {k: v[0, 0] for k, v in tb.items()}

    return tb, coords


def build_discretized(tb_hamiltonian, coords, *, grid=None, locals=None,
                      grid_spacing=None):
    """Create a template builder from a symbolic tight-binding Hamiltonian.

    The provided symbolic tight-binding Hamiltonian is put on a (hyper) square
    lattice and turned into Python functions.  These functions are used to
    create a `~kwant.builder.Builder` instance that may be used with
    `~kwant.builder.Builder.fill` to construct a system of a desired shape.

    The return values of `~kwant.continuum.discretize_symbolic` may be used
    directly for the first two arguments of this function.

    .. warning::
        This function uses ``eval`` (because it calls ``sympy.sympify``), and
        thus should not be used on unsanitized input.

    Parameters
    ----------
    tb_hamiltonian : dict
        Keys must be the offsets of the hoppings, represented by tuples of
        integers ((0, 0, 0) for onsite). Values must be symbolic expressions
        for the hoppings/onsite or expressions that can by sympified with
        `kwant.continuum.sympify`.
    coords : sequence of strings
        The coordinates for which momentum operators will be treated as
        differential operators. May contain only "x", "y" and "z" and must be
        sorted.
    grid : scalar or kwant.lattice.Monatomic instance, optional
        Lattice that will be used as a discretization grid. It must have
        orthogonal primitive vectors. If a scalar value is given, a lattice
        with the appropriate grid spacing will be generated. If not provided,
        a lattice with grid spacing 1 in all directions will be generated.
    locals : dict, optional
        Additional namespace entries for `~kwant.continuum.sympify`.  May be
        used to simplify input of matrices or modify input before proceeding
        further. For example:
        ``locals={'k': 'k_x + I * k_y'}`` or
        ``locals={'sigma_plus': [[0, 2], [0, 0]]}``.
    grid_spacing : int or float, optional
        (deprecated) Spacing of the discretization grid. If not provided,
        the default value will be 1. Cannot be used together with ``grid``.

    Returns
    -------
    model : `~kwant.builder.Builder`
        The translationally symmetric builder that corresponds to the provided
        Hamiltonian.  This builder instance belongs to a subclass of the
        standard builder that may be printed to obtain the source code of the
        value functions.  It also holds the discretization lattice (a
        `~kwant.lattice.Monatomic` instance with lattice constant
        `grid_spacing`) in the ``lattice`` attribute.
    """
    # check already available constraints (grid will be check later)
    if len(coords) == 0:
        raise ValueError('Discrete coordinates cannot be empty.')

    if grid_spacing is not None:  # TODO: remove when we remove 'grid_spacing'
        warnings.warn('The "grid_spacing" parameter is deprecated. Use '
                      '"grid" instead.', KwantDeprecationWarning, stacklevel=3)
    if grid is None and grid_spacing is None:
        grid = 1  # default case
    elif grid is None:  # TODO: remove when we remove 'grid_spacing'
        grid = grid_spacing
    elif grid_spacing is not None:
        raise ValueError('"grid_spacing" and "grid" are mutually exclusive.')

    coords = list(coords)
    grid_dim = len(coords)

    if coords != sorted(coords):
        raise ValueError("The argument 'coords' must be sorted.")

    # run sympifcation on hamiltonian values
    with reraise_warnings():
        for k, v in tb_hamiltonian.items():
            tb_hamiltonian[k] = sympify(v, locals)

    # generate grid if required, check constraints if provided
    random_element = next(iter(tb_hamiltonian.values()))
    norbs = (1 if isinstance(random_element, sympy.Expr)
             else random_element.shape[0])

    if np.isscalar(grid):
        lat = lattice.Monatomic(grid * np.eye(grid_dim), norbs=norbs)
    else:
        lat = grid

    # check grid constraints
    is_diagonal = lambda m: np.allclose(m, np.diag(np.diagonal(m)))
    if not (lat.prim_vecs.shape[0] == grid_dim and
            is_diagonal(lat.prim_vecs)):
        raise ValueError('"grid" is expected to by an orthogonal lattice '
                         'of dimension matching number of "coords".')

    if (lat.norbs is not None) and (lat.norbs != norbs):
        raise ValueError(
            'Number of lattice orbitals does not match the number '
            'of orbitals in the Hamiltonian.'
        )

    # continue with building the template
    tb = {}
    for n, (offset, hopping) in enumerate(tb_hamiltonian.items()):
        onsite = all(i == 0 for i in offset)

        if onsite:
            name = 'onsite'
        else:
            name = 'hopping_{}'.format(n)

        tb[offset] = _builder_value(hopping, coords, np.diag(lat.prim_vecs),
                                    onsite, name)

    onsite_zeros = (0,) * grid_dim
    onsite = tb.pop(onsite_zeros)
    # 'delta' parameter to HoppingKind is the negative of the 'hopping offset'
    hoppings = {builder.HoppingKind(tuple(-i for i in d), lat): val
                for d, val in tb.items()}

    syst = _DiscretizedBuilder(
        coords, lat, lattice.TranslationalSymmetry(*lat.prim_vecs)
    )
    syst[lat(*onsite_zeros)] = onsite
    for hop, val in hoppings.items():
        syst[hop] = val

    return syst


def _differentiate(expression, coordinate_name):
    """Calculate derivative of an expression for given coordinate.

    Parameters:
    -----------
    expression : sympy.Expr
        Sympy expression containing function to be derivated.
    coordinate_name : string
        Coordinate over which derivative is calculated.

    Returns
    -------
    sympy.Expr
    """
    assert coordinate_name in 'xyz'
    coordinate = _position_operators[coordinate_name]
    h = _displacements[coordinate_name]

    expr1 = expression.subs(coordinate, coordinate + h)
    expr2 = expression.subs(coordinate, coordinate - h)

    return (expr1 - expr2) / (2 * h)


def _discretize_summand(summand, coords):
    """Discretize a product of factors.

    Parameters
    ----------
    summand : sympy.Expr
    coords : sequence of strings
        Must be a subset of ``{'x', 'y', 'z'}``.

    Returns
    -------
    sympy.Expr
    """
    assert not isinstance(summand, sympy.Add), "Input should be one summand."
    momenta = ['k_{}'.format(s) for s in coords]

    factors = reversed(summand.as_ordered_factors())
    result = 1
    for factor in factors:
        symbol, exponent = factor.as_base_exp()
        if isinstance(symbol, sympy.Symbol) and symbol.name in momenta:
            for i in range(exponent):
                coordinate = symbol.name[-1]
                # apply momentum as differential operator '-i d/dx'
                result = -sympy.I * _differentiate(result, coordinate)
        else:
            result = factor * result

    return result


def _discretize_expression(expression, coords):
    """Discretize an expression into a discrete (tight-binding) representation.

    Parameters
    ----------
    expression : sympy.Expr
    coords : sequence of strings
        Must be a subset of ``{'x', 'y', 'z'}``.

    Returns
    -------
    dict
        Keys are tuples of integers; the offsets of the hoppings
        ((0, 0, 0) for the onsite). Values are symbolic expressions
        for the hoppings/onsite.
    """
    def _read_offset(wf):
        # e.g. wf(x + h, y, z + h) -> (1, 0, 1)
        assert wf.func == _wf

        offset = []
        for c, arg in zip(coords, wf.args):
            coefficients = arg.as_coefficients_dict()
            assert coefficients[_position_operators[c]] == 1

            ai = _displacements[c]
            offset.append(coefficients.pop(ai, 0))
        return tuple(offset)

    def _extract_hoppings(expr):
        """Read hoppings and perform shortening operation."""
        expr = sympy.expand(expr)
        summands = [e.as_ordered_factors() for e in expr.as_ordered_terms()]

        offset = [_read_offset(s[-1]) for s in summands]
        coeffs = [sympy.Mul(*s[:-1]) for s in summands]
        offset = np.array(offset, dtype=int)
        # rescale the offsets for each coordinate by their greatest
        # common divisor across the summands. e.g:
        # wf(x+2h) + wf(x+4h) --> wf(x+h) + wf(x+2h) and a_x //= 2
        subs = {}
        for i, xi in enumerate(coords):
            factor = int(gcd(*offset[:, i]))
            if factor < 1:
                continue
            offset[:, i] //= factor
            subs[_displacements[xi]] = _displacements[xi] / factor
        # apply the rescaling to the hoppings
        output = defaultdict(lambda: sympy.Integer(0))
        for n, c in enumerate(coeffs):
            output[tuple(offset[n].tolist())] += c.subs(subs)
        return dict(output)

    # if there are no momenta in the expression, then it is an onsite
    atoms_names = [s.name for s in expression.atoms(sympy.Symbol)]
    if not set(_momentum_operators) & set(atoms_names):
        n = len(coords)
        return {(0,) * n: expression}

    # make sure we have list of summands
    summands = expression.as_ordered_terms()

    # discretize every summand
    coordinates = tuple(_position_operators[s] for s in coords)
    wf = _wf(*coordinates)

    discrete_expression = defaultdict(int)
    for summand in summands:
        summand = _discretize_summand(summand * wf, coords)
        hops = _extract_hoppings(summand)
        for k, v in hops.items():
            discrete_expression[k] += v

    return dict(discrete_expression)


################ string processing

class _NumericPrinter(LambdaPrinter):

    def __init__(self):
        if 'allow_unknown_functions' in LambdaPrinter._default_settings:
            settings = {'allow_unknown_functions': True}
        else:
            # We're on Sympy without "allow_unknown_functions" setting
            settings = {}

        LambdaPrinter.__init__(self, settings=settings)

        self.known_functions = {}
        self.known_constants = {'pi': 'pi', 'Pi': 'pi', 'I': 'I'}

    def _print_ImaginaryUnit(self, expr):
        # prevent sympy from printing 'I' for imaginary unit
        return "1j"

    def _print_Pow(self, expr):
        # copied from sympy's StrPrinter with the code paths
        # to print 'sqrt' removed.
        PREC = precedence(expr)

        if expr.is_commutative and expr.exp is -sympy.S.One:
            return '1/%s' % self.parenthesize(expr.base, PREC)

        e = self.parenthesize(expr.exp, PREC)
        if (self.printmethod == '_sympyrepr' and
            expr.exp.is_Rational and expr.exp.q != 1):
            if e.startswith('(Rational'):
                return '%s**%s' % (self.parenthesize(expr.base, PREC), e[1:-1])
        return '%s**%s' % (self.parenthesize(expr.base, PREC), e)


def _print_sympy(expr):
    return lambdastr((), expr, printer=_NumericPrinter)[len('lambda : '):]


def _return_string(expr, coords):
    """Process a sympy expression into an evaluatable Python return statement.

    Parameters
    ----------
    expr : sympy.Expr

    Returns
    -------
    output : string
        A return string that can be used to assemble a Kwant value function.
    map_func_calls : dict
        mapping of function calls to assigned constants.
    const_symbols : sequance of sympy.Symbol
        All constants that appear in the expression.
    _cache: dict
        mapping of cache symbols to cached matrices.
    """
    _cache = {}
    def cache(x):
        s = sympy.symbols('_cache_{}'.format(len(_cache)))
        _cache[str(s)] = ta.array(x.tolist(), complex)
        return s

    blacklisted = set(coords) | {'site', 'site1', 'site2'}
    const_symbols = {s for s in expr.atoms(sympy.Symbol)
                     if s.name not in blacklisted}

    # functions will be evaluated within the function body and the
    # result assigned to a symbol '_const_<n>', so we replace all
    # function calls by these symbols in the return statement.
    map_func_calls = expr.atoms(AppliedUndef, sympy.Function)
    map_func_calls = {s: sympy.symbols('_const_{}'.format(n))
                      for n, s in enumerate(map_func_calls)}

    expr = expr.subs(map_func_calls)

    if isinstance(expr, MatrixBase):
        # express matrix return values in terms of sums of known matrices,
        # which will be assigned to '_cache_n' in the function body.
        mons = monomials(expr, expr.atoms(sympy.Symbol))
        mons = {k: cache(v) for k, v in mons.items()}
        mons = ["{} * {}".format(_print_sympy(k), _print_sympy(v))
                for k, v in mons.items()]
        output = " + ".join(mons)
    else:
        output = _print_sympy(expr)

    return 'return {}'.format(output), map_func_calls, const_symbols, _cache


def _assign_symbols(map_func_calls, coords, onsite):
    """Generate a series of assignments.

    Parameters
    ----------
    map_func_calls : dict
        mapping of function calls to assigned constants.
    coords : sequence of strings
        If left as None coordinates will not be read from a site.
    onsite : bool
        True if function is called for onsite, false for hoppings

    Returns
    -------
    assignments : list of strings
        List of lines used for including in a function.
    """
    lines = []

    if coords:
        site = 'site' if onsite else 'site1'
        args = ', '.join(coords), site
        lines.append('({}, ) = {}.pos'.format(*args))

    for k, v in map_func_calls.items():
        lines.append("{} = {}".format(v, _print_sympy(k)))

    return lines


def _builder_value(expr, coords, grid_spacing, onsite,
                   name='_anonymous_func'):
    """Generate a builder value from a sympy expression.

    Parameters
    ----------
    expr : sympy.Expr or sympy.matrix
        Expr that from which value function will be generated.
    coords : sequence of strings
        List of coodinates present in the system.
    grid_spacing : sequence of scalars
        Lattice spacing of the system in each coordinate.

    Returns
    -------
    `expr` transformed into an object that can be used as a
    `kwant.builder.Builder` value.  Either a numerical value
    (``tinyarray.array`` instance or complex number) or a value function.  In
    the case of a function, the source code is available in its `_source`
    attribute.
    """

    expr = expr.subs({_displacements[c]: grid_spacing[n]
                      for n, c in enumerate(coords)})
    return_string, map_func_calls, const_symbols, _cache = _return_string(
        expr, coords=coords)

    # first check if value function needs to read coordinates
    atoms_names = {s.name for s in expr.atoms(sympy.Symbol)}
    if not set(coords) & atoms_names:
        coords = None

    # constants and functions in the sympy input will be passed
    # as arguments to the value function
    arg_names = set.union({s.name for s in const_symbols},
                          {str(k.func) for k in map_func_calls})

    # check if all argument names are valid python identifiers
    for arg_name in arg_names:
        if not (arg_name.isidentifier() and not iskeyword(arg_name)):
            raise ValueError("Invalid name in used symbols: {}\n"
                             "Names of symbols used in Hamiltonian "
                             "must be valid Python identifiers and "
                             "may not be keywords".format(arg_name))

    arg_names = ', '.join(sorted(arg_names))

    if (not arg_names) and (coords is None):
        # we can just use a constant value instead of a value function
        if isinstance(expr, sympy.MatrixBase):
            return ta.array(expr.tolist(), complex)
        else:
            return complex(expr)

    lines = _assign_symbols(map_func_calls, onsite=onsite, coords=coords)
    lines.append(return_string)

    separator = '\n    '
    # 'site_string' is tightly coupled to the symbols used in '_assign_symbol'
    site_string = 'site' if onsite else 'site1, site2'
    if arg_names:
        header_str = 'def {}({}, {}):'
        header = header_str.format(name, site_string, arg_names)
    else:
        header = 'def {}({}):'.format(name, site_string)
    func_code = separator.join([header] + list(lines))

    # Add "I" to namespace just in case sympy again would miss to replace it
    # with Python's 1j as it was the case with SymPy 1.2 when I was argument
    # of some function.
    namespace = {'pi': np.pi, 'I': 1j}
    namespace.update(_cache)

    source = []
    for k, v in _cache.items():
        source.append("{} = (\n{})\n".format(k, repr(np.array(v))))
    source.append(func_code)

    exec(func_code, namespace)
    f = namespace[name]
    f._source = "".join(source)

    return f



# Copyright 2011-2018 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.
"""Functions for fixing the magnetic gauge automatically in a Kwant system.


The "gauge" module has been included in Kwant on a provisional basis.
Backwards incompatible changes (up to and including removal of the package)
may occur if deemed necessary by the core developers.
"""

import bisect
from functools import partial
from itertools import permutations

import numpy as np
import scipy
from scipy.integrate import dblquad
from scipy.sparse import csgraph

from .. import system, builder

from ..graph.dijkstra import dijkstra_directed

__all__ = ['magnetic_gauge']


### Integation

# Integrate vector field over triangle, for internal use by '_surface_integral'
# Triangle is (origin, origin + v1, origin + v2), 'n' is np.cross(v1, v2)

def _quad_triangle(f, origin, v1, v2, n, tol):
    if np.dot(n, n) < tol**2:  # does triangle have significant area?
        return 0

    def g(x, y):
        return np.dot(n, f(origin + x * v1 + y * v2))

    result, *_ = dblquad(g, 0, 1, lambda x: 0, lambda x: 1-x)
    return result.real


def _const_triangle(f, origin, v1, v2, n, tol):
    return np.dot(f, n) / 2


def _average_triangle(f, origin, v1, v2, n, tol):
    return np.dot(n, f(origin + 1/3 * (v1 + v2))) / 2


def _surface_integral(f, loop, tol=1e-8, average=False):
    """Calculate the surface integral of 'f' over a surface enclosed by 'loop'.

    This function only works for *divergence free* vector fields, where the
    surface integral depends only on the boundary.

    Parameters
    ----------
    f : callable or real n-vector
        The vector field for which to calculate the surface integral.
        If callable, takes a real n-vector as argument and returns a
        real n-vector.
    loop : sequence of vectors
        Ordered sequence of real n-vectors (positions) that define the
        vertices of the polygon that encloses the surface to integrate over.
    tol : float, default: 1e-8
        Error tolerance on the result.
    average : bool, default: False
        If True, approximate the integral over each triangle using a
        single function evaluation at the centre of the triangle.
    """
    if callable(f):
        integrator = _average_triangle if average else _quad_triangle
    else:
        integrator = _const_triangle

    origin, *points = loop
    integral = 0
    # Split loop into triangles with 1 vertex on 'origin', evaluate
    # the integral over each triangle and sum the result
    for p1, p2 in zip(points, points[1:]):
        v1 = p1 - origin
        v2 = p2 - origin
        n = np.cross(v1, v2)
        integral += integrator(f, origin, v1, v2, n, tol)
    return integral


### Loop finding graph algorithm

def _find_loops(graph, subgraph):
    """
    Parameters
    ----------
    graph : COO matrix
        The complete undirected graph, where the values of the matrix are
        the weights of the corresponding graph links.
    subgraph : COO matrix
        An subgraph of 'graph', with missing edges denoted by infinities.
        Must have the same sparsity structure as 'graph'.

    Returns
    -------
    A sequence of paths which are partially contained in the subgraph.
    The loop is formed by adding a link between the first and last node.

    The ordering is such that the paths are made of links that belong to
    the subgraph or to the previous closed loops.
    """
    # For each link we do 1 update of 'subgraph' and a call to
    # 'csgraph.shortest_path'. It is cheaper to update the CSR
    # matrix rather than convert to LIL and back every iteration.
    subgraph = subgraph.tocsr()
    graph = graph.tocsr()
    assert _same_sparsity_structure(subgraph, graph)

    # Links in graph, but not in subgraph.
    links_to_find = scipy.sparse.triu(graph - subgraph).tocoo()
    links_to_find = np.vstack((links_to_find.row, links_to_find.col)).transpose()

    links_to_find, min_length = _order_links(subgraph, links_to_find)

    # Find shortest path between each link in turn, updating the subgraph with
    # the links as we go.
    loops = []
    while len(links_to_find) > 0:
        frm, to = links_to_find[0]
        (path,), (path_length,) = dijkstra_directed(subgraph,
                                                    sources=np.array([frm]),
                                                    targets=np.array([to]))

        # Reorder links that are still to find based on the loop length in
        # the updated graph. We only reorder when the path length for *this*
        # link is a "little bit" longer that the perviously determined minimum.
        # The "little bit" is needed so we don't needlessly re-order the links
        # on amorphous lattices.
        if path_length > min_length * 1.1:
            links_to_find, min_length = _order_links(subgraph, links_to_find)
        else:
            # Assumes that 'graph' and 'subgraph' have the same sparsity structure.
            _assign_csr(subgraph, graph, (frm, to))
            _assign_csr(subgraph, graph, (to, frm))
            loops.append(path)
            links_to_find = links_to_find[1:]

    return loops


def _order_links(subgraph, links_to_find):
    if len(links_to_find) == 0:
        return [], None
    # Order 'links_to_find' by length of shortest path between the nodes of the link
    path_lengths = dijkstra_directed(subgraph,
                                     sources=links_to_find[:, 0],
                                     targets=links_to_find[:, 1],
                                     return_paths=False)
    idxs = np.argsort(path_lengths)
    return links_to_find[idxs], path_lengths[idxs[0]]


### Generic sparse matrix utilities

def _assign_csr(a, b, element):
    """Assign a single element from a CSR matrix to another.

    Parameters
    ----------
    a : CSR matrix
    b : CSR matrix or scalar
        If a CSR matrix, must have the same sparsity structure
        as 'a'. If a scalar, must be the same dtype as 'a'.
    element: (int, int)
        Row and column indices of the element to set.
    """
    assert isinstance(a, scipy.sparse.csr_matrix)
    row, col = element
    for j in range(a.indptr[row], a.indptr[row + 1]):
        if a.indices[j] == col:
            break
    else:
        raise ValueError('{} not in sparse matrix'.format(element))
    if isinstance(b, scipy.sparse.csr_matrix):
        a.data[j] = b.data[j]
    else:
        a.data[j] = b


def _same_sparsity_structure(a, b):
    a = a.tocsr().sorted_indices()
    b = b.tocsr().sorted_indices()
    return (np.array_equal(a.indices, b.indices)
            and np.array_equal(a.indptr, b.indptr))


def _add_coo_matrices(*mats, shape):
    """Add a sequence of COO matrices by appending their constituent arrays."""
    values = np.hstack([mat.data for mat in mats])
    rows = np.hstack([mat.row for mat in mats])
    cols = np.hstack([mat.col for mat in mats])
    return scipy.sparse.coo_matrix((values, (rows, cols)), shape=shape)


def _shift_diagonally(mat, shift, shape):
    """Shift the row/column indices of a COO matrix."""
    return scipy.sparse.coo_matrix(
        (mat.data, (mat.row + shift, mat.col + shift)),
        shape=shape)


def _distance_matrix(links, pos, shape):
    """Return the distances between the provided links as a COO matrix.

    Parameters
    ----------
    links : sequence of pairs of int
        The links for which to find the lengths.
    pos : callable: int -> vector
        Map from link ends (integers) to realspace position.
    shape : tuple
    """
    if len(links) == 0:  # numpy does not like 'if array'
        return scipy.sparse.coo_matrix(shape)
    links = np.array(links)
    distances = np.array([pos(i) - pos(j) for i, j in links])
    distances = np.linalg.norm(distances, axis=1)
    return scipy.sparse.coo_matrix((distances, links.T), shape=shape)


### Loop finding
#
# All of these functions take a finalized Kwant system and return
# a sequence of loops. Each loop is a sequence of sites (integers)
# that one visits when traversing the loop. The first and final sites
# are assumed to be linked, which closes the loop. The links that one
# traverses when going round a loop is thus:
#
#    list(zip(loop, loop[1:])) + [(loop[-1], loop[0])]
#
# These loops are later used to fix the magnetic gauge in the system.
# All of the links except the final one are assumed to have their gauge
# fixed (i.e. the phase across them is known), and gauge of the final
# link is the one to be determined.


def _loops_in_finite(syst):
    """Find the loops in a finite system with no leads.

    The site indices in the returned loops are those of the system,
    so they may be used as indices to 'syst.sites', or with 'syst.pos'.
    """
    assert isinstance(syst, system.FiniteSystem) and syst.leads == []
    nsites = len(syst.sites)

    # Fix the gauge across the minimum spanning tree of the system graph.
    graph = _distance_matrix(list(syst.graph),
                             pos=syst.pos, shape=(nsites, nsites))
    spanning_tree = _shortest_distance_forest(graph)
    return _find_loops(graph, spanning_tree)


def _shortest_distance_forest(graph):
    # Grow a forest of minimum distance trees for all connected components of the graph
    graph = graph.tocsr()
    tree = graph.copy()
    # set every entry in tree to infinity
    tree.data[:] = np.inf
    unvisited = set(range(graph.shape[0]))
    # set the target node to be greater than any node in the graph.
    # This way we explore the whole graph.
    end = np.array([graph.shape[0] + 1], dtype=np.int32)

    while unvisited:
        # Choose an arbitrary element as the root
        root = unvisited.pop()
        root = np.array([root], dtype=np.int32)
        _, pred = dijkstra_directed(graph, sources=root, targets=end,
                                    return_predecessors=True, return_paths=False)
        for i, p in enumerate(pred):
            # -1 if node 'i' has no predecessor. Either it is the root node,
            # or it was not reached.
            if p != -1:
                unvisited.remove(i)
                _assign_csr(tree, graph, (i, p))
                _assign_csr(tree, graph, (p, i))
    return tree


def _loops_in_infinite(syst):
    """Find the loops in an infinite system.

    Returns
    -------
    loops : sequence of sequences of integers
        The sites in the returned loops belong to two adjacent unit
        cells. The first 'syst.cell_size' sites are in the first
        unit cell, and the next 'sys.cell_size' are in the next
        (in the direction of the translational symmetry).
    extended_sites : callable : int -> Site
        Given a site index in the extended system consisting of
        two unit cells, returns the associated high-level
        `kwant.builder.Site`.
    """
    assert isinstance(syst, system.InfiniteSystem)
    _check_infinite_syst(syst)

    cell_size = syst.cell_size

    unit_cell_links = [(i, j) for i, j in syst.graph
                       if i < cell_size and j < cell_size]
    unit_cell_graph = _distance_matrix(unit_cell_links,
                                       pos=syst.pos,
                                       shape=(cell_size, cell_size))

    # Loops in the interior of the unit cell
    spanning_tree = _shortest_distance_forest(unit_cell_graph)
    loops = _find_loops(unit_cell_graph, spanning_tree)

    # Construct an extended graph consisting of 2 unit cells connected
    # by the inter-cell links.
    extended_shape = (2 * cell_size, 2 * cell_size)
    uc1 = _shift_diagonally(unit_cell_graph, 0, shape=extended_shape)
    uc2 = _shift_diagonally(unit_cell_graph, cell_size, shape=extended_shape)
    hop_links = [(i, j) for i, j in syst.graph if j >= cell_size]
    hop = _distance_matrix(hop_links,
                           pos=syst.pos,
                           shape=extended_shape)
    graph = _add_coo_matrices(uc1, uc2, hop, hop.T,
                              shape=extended_shape)

    # Construct a subgraph where only the shortest link between the
    # 2 unit cells is added. The other links are added with infinite
    # values, so that the subgraph has the same sparsity structure
    # as 'graph'.
    idx = np.argmin(hop.data)
    data = np.full_like(hop.data, np.inf)
    data[idx] = hop.data[idx]
    smallest_edge = scipy.sparse.coo_matrix(
        (data, (hop.row, hop.col)),
        shape=extended_shape)
    subgraph = _add_coo_matrices(uc1, uc2, smallest_edge, smallest_edge.T,
                                 shape=extended_shape)

    # Use these two graphs to find the loops between unit cells.
    loops.extend(_find_loops(graph, subgraph))

    def extended_sites(i):
        unit_cell = np.array([i // cell_size])
        site = syst.sites[i % cell_size]
        return syst.symmetry.act(-unit_cell, site)

    return loops, extended_sites


def _loops_in_composite(syst):
    """Find the loops in finite system with leads.

    Parameters
    ----------
    syst : kwant.builder.FiniteSystem

    Returns
    -------
    loops : sequence of sequences of integers
        The sites in each loop belong to the extended scattering region
        (see notes). The first and last site in each loop are guaranteed
        to be in 'syst'.
    which_patch : callable : int -> int
        Given a site index in the extended scattering region (see notes),
        returns the lead patch (see notes) to which the site belongs. Returns
        -1 if the site is part of the reduced scattering region (see notes).
    extended_sites : callable : int -> Site
        Given a site index in the extended scattering region (see notes),
        returns the associated high-level `kwant.builder.Site`.

    Notes
    -----
    extended scattering region
        The scattering region with a single lead unit cell attached at
        each interface. This unit cell is added so that we can "see" any
        loops formed with sites in the lead (see 'check_infinite_syst'
        for details). The sites for each lead are added in the same
        order as the leads, and within a given added unit cell the sites
        are ordered in the same way as the associated lead.
    lead patch
        Sites in the extended scattering region that belong to the added
        unit cell for a given lead, or the lead padding for a given lead
        are said to be in the "lead patch" for that lead.
    reduced scattering region
        The sites of the extended scattering region that are not
        in a lead patch.
    """
    # Check that we can consistently fix the gauge in the scattering region,
    # given that we have independently fixed gauges in the leads.
    _check_composite_system(syst)

    # Get distance matrix for the extended scattering region,
    # a function that maps sites to their lead patches (-1 for sites
    # in the reduced scattering region), and a function that maps sites
    # to high-level 'kwant.builder.Site' objects.
    distance_matrix, which_patch, extended_sites =\
        _extended_scattering_region(syst)

    spanning_tree = _spanning_tree_composite(distance_matrix, which_patch).tocsr()

    # Fill in all links with at least 1 site in a lead patch;
    # their gauge is fixed by the lead gauge.
    for i, j, v in zip(distance_matrix.row, distance_matrix.col,
                       distance_matrix.data):
        if which_patch(i) > -1 or which_patch(j) > -1:
            _assign_csr(spanning_tree, v, (i, j))
            _assign_csr(spanning_tree, v, (j, i))

    loops = _find_loops(distance_matrix, spanning_tree)

    return loops, which_patch, extended_sites


def _extended_scattering_region(syst):
    """Return the distance matrix of a finite system with 1 unit cell
       added to each lead interface.

    Parameters
    ----------
    syst : kwant.builder.FiniteSystem

    Returns
    -------
    extended_scattering_region: COO matrix
        Distance matrix between connected sites in the extended
        scattering region.
    which_patch : callable : int -> int
        Given a site index in the extended scattering region, returns
        the lead patch to which the site belongs. Returns
        -1 if the site is part of the reduced scattering region.
    extended_sites : callable : int -> Site
        Given a site index in the extended scattering region, returns
        the associated high-level `kwant.builder.Site`.

    Notes
    -----
    Definitions of the terms 'extended scatteringr region',
    'lead patch' and 'reduced scattering region' are given
    in the notes for `kwant.physics.gauge._loops_in_composite`.
    """
    extended_size = (syst.graph.num_nodes
                     + sum(l.cell_size for l in syst.leads))
    extended_shape = (extended_size, extended_size)

    added_unit_cells = []
    first_lead_site = syst.graph.num_nodes
    for lead, interface in zip(syst.leads, syst.lead_interfaces):
        # Here we assume that the distance between sites in the added
        # unit cell and sites in the interface is the same as between sites
        # in neighboring unit cells.
        uc = _distance_matrix(list(lead.graph),
                              pos=lead.pos, shape=extended_shape)
        # Map unit cell lead sites to their indices in the extended scattering,
        # region and sites in next unit cell to their interface sites.
        hop_from_syst = uc.row >= lead.cell_size
        uc.row[~hop_from_syst] = uc.row[~hop_from_syst] + first_lead_site
        uc.row[hop_from_syst] = interface[uc.row[hop_from_syst] - lead.cell_size]
        # Same for columns
        hop_to_syst = uc.col >= lead.cell_size
        uc.col[~hop_to_syst] = uc.col[~hop_to_syst] + first_lead_site
        uc.col[hop_to_syst] = interface[uc.col[hop_to_syst] - lead.cell_size]

        added_unit_cells.append(uc)
        first_lead_site += lead.cell_size

    scattering_region = _distance_matrix(list(syst.graph),
                                         pos=syst.pos, shape=extended_shape)

    extended_scattering_region = _add_coo_matrices(scattering_region,
                                                   *added_unit_cells,
                                                   shape=extended_shape)

    lead_starts = np.cumsum([syst.graph.num_nodes,
                             *[lead.cell_size for lead in syst.leads]])
    # Frozenset to quickly check 'is this site in the lead padding?'
    extra_sites = [frozenset(sites) for sites in syst.lead_paddings]


    def which_patch(i):
        if i < len(syst.sites):
            # In scattering region
            for patch_num, sites in enumerate(extra_sites):
                if i in sites:
                    return patch_num
            # If not in 'extra_sites' it is in the reduced scattering region.
            return -1
        else:
            # Otherwise it's in an attached lead cell
            which_lead = bisect.bisect(lead_starts, i) - 1
            assert which_lead > -1
            return which_lead


    # Here we use the fact that all the sites in a lead interface belong
    # to the same symmetry domain.
    interface_domains = [lead.symmetry.which(syst.sites[interface[0]])
                         for lead, interface in
                         zip(syst.leads, syst.lead_interfaces)]

    def extended_sites(i):
        if i < len(syst.sites):
            # In scattering region
            return syst.sites[i]
        else:
            # Otherwise it's in an attached lead cell
            which_lead = bisect.bisect(lead_starts, i) - 1
            assert which_lead > -1
            lead = syst.leads[which_lead]
            domain = interface_domains[which_lead] + 1
            # Map extended scattering region site index to site index in lead.
            i = i - lead_starts[which_lead]
            return lead.symmetry.act(domain, lead.sites[i])

    return extended_scattering_region, which_patch, extended_sites


def _interior_links(distance_matrix, which_patch):
    """Return the indices of the links in 'distance_matrix' that
       connect interface sites of the scattering region to other
       sites (interface and non-interface) in the scattering region.
    """

    def _is_in_lead(i):
        return which_patch(i) > -1

    # Sites that connect to/from sites in a lead patch
    interface_sites = {
        (i if not _is_in_lead(i) else j)
        for i, j in zip(distance_matrix.row, distance_matrix.col)
        if _is_in_lead(i) ^ _is_in_lead(j)
    }

    def _we_want(i, j):
        return i in interface_sites and not _is_in_lead(j)

    # Links that connect interface sites to the rest of the scattering region.
    return np.array([
        k
        for k, (i, j) in enumerate(zip(distance_matrix.row, distance_matrix.col))
        if _we_want(i, j) or _we_want(j, i)
    ])


def _make_metatree(graph, links_to_delete):
    """Make a tree of the components of 'graph' that are
       disconnected by deleting 'links'. The values of
       the returned tree are indices of edges in 'graph'
       that connect components.
    """
    # Partition the graph into disconnected components
    dl = partial(np.delete, obj=links_to_delete)
    partitioned_graph = scipy.sparse.coo_matrix(
        (dl(graph.data), (dl(graph.row), dl(graph.col)))
    )
    # Construct the "metagraph", where each component is reduced to
    # a single node, and a representative (smallest) edge is chosen
    # among the edges that connected the components in the original graph.
    ncc, labels = csgraph.connected_components(partitioned_graph)
    metagraph = scipy.sparse.dok_matrix((ncc, ncc), int)
    for k in links_to_delete:
        i, j = labels[graph.row[k]], labels[graph.col[k]]
        if i == j:
            continue  # Discard loop edges
        # Add a representative (smallest) edge from each graph component.
        if graph.data[k] < metagraph.get((i, j), np.inf):
            metagraph[i, j] = k
            metagraph[j, i] = k

    return csgraph.minimum_spanning_tree(metagraph).astype(int)


def _spanning_tree_composite(distance_matrix, which_patch):
    """Find a spanning tree for a composite system.

    We cannot use a simple minimum-distance spanning tree because
    we have the additional constraint that all links with at least
    one end in a lead patch have their gauge fixed. See the notes
    for details.

    Parameters
    ----------
    distance_matrix : COO matrix
        Distance matrix between connected sites in the extended
        scattering region.
    which_patch : callable : int -> int
        Given a site index in the extended scattering region (see notes),
        returns the lead patch (see notes) to which the site belongs. Returns
        -1 if the site is part of the reduced scattering region (see notes).
    Returns
    -------
    spanning_tree : CSR matrix
        A spanning tree with the same sparsity structure as 'distance_matrix',
        where missing links are denoted with infinite weights.

    Notes
    -----
    Definitions of the terms 'extended scattering region', 'lead patch'
    and 'reduced scattering region' are given in the notes for
    `kwant.physics.gauge._loops_in_composite`.

    We cannot use a simple minimum-distance spanning tree because
    we have the additional constraint that all links with at least
    one end in a lead patch have their gauge fixed.
    Consider the following case using a minimum-distance tree
    where 'x' are sites in the lead patch::

        o-o-x      o-o-x
        | | |  -->   | |
        o-o-x      o-o x

    The removed link on the lower right comes from the lead, and hence
    is gauge-fixed, however the vertical link in the center is not in
    the lead, but *is* in the tree, which means that we will fix its
    gauge to 0. The loop on the right would thus not have the correct
    gauge on all links.

    Instead we first cut all links between *interface* sites and
    sites in the scattering region (including other interface sites).
    We then construct a minimum distance forest for these disconnected
    graphs. Finally we add back links from the ones that were cut,
    ensuring that we do not form any loops; we do this by contructing
    a tree of representative links from the "metagraph" of components
    that were disconnected by the link cutting.
    """
    # Links that connect interface sites to other sites in the
    # scattering region (including other interface sites)
    links_to_delete = _interior_links(distance_matrix, which_patch)
    # Make a shortest distance tree for each of the components
    # obtained by cutting the links.
    cut_syst = distance_matrix.copy()
    cut_syst.data[links_to_delete] = np.inf
    forest = _shortest_distance_forest(cut_syst)
    # Connect the forest back up with representative links until
    # we have a single tree (if the original system was not connected,
    # we get a forest).
    metatree = _make_metatree(distance_matrix, links_to_delete)
    for k in np.unique(metatree.data):
        value = distance_matrix.data[k]
        i, j = distance_matrix.row[k], distance_matrix.col[k]
        _assign_csr(forest, value, (i, j))
        _assign_csr(forest, value, (j, i))

    return forest


def _check_infinite_syst(syst):
    r"""Check that the unit cell is a connected graph.

    If the unit cell is not connected then we cannot be sure whether
    there are loops or not just by inspecting the unit cell graph
    (this may be a solved problem, but we could not find an algorithm
    to do this).

    To illustrate this, consider the following unit cell consisting
    of 3 sites and 4 hoppings::

        o-
         \
        o
         \
        o-

    None of the sites are connected within the unit cell, however if we repeat
    a few unit cells::

        o-o-o-o
         \ \ \
        o o o o
         \ \ \
        o-o-o-o

    we see that there is a loop crossing 4 unit cells. A connected unit cell
    is a sufficient condition that all the loops can be found by inspecting
    the graph consisting of two unit cells glued together.
    """
    assert isinstance(syst, system.InfiniteSystem)
    n = syst.cell_size
    rows, cols = np.array([(i, j) for i, j in syst.graph
                            if i < n and j < n]).transpose()
    data = np.ones(len(rows))
    graph = scipy.sparse.coo_matrix((data, (rows, cols)), shape=(n, n))
    if csgraph.connected_components(graph, return_labels=False) > 1:
        raise ValueError(
            'Infinite system unit cell is not connected: we cannot determine '
            'if there are any loops in the system\n\n'
            'If there are, then you must define your unit cell so that it is '
            'connected. If there are not, then you must add zero-magnitude '
            'hoppings to your system.'
        )


def _check_composite_system(syst):
    """Check that we can consistently fix the gauge in a system with leads.

    If not, raise an exception with an informative error message.
    """
    assert isinstance(syst, system.FiniteSystem) and syst.leads
    # Frozenset to quickly check 'is this site in the lead padding?'
    extras = [frozenset(sites) for sites in syst.lead_paddings]
    interfaces = [set(iface) for iface in syst.lead_interfaces]
    # Make interfaces between lead patches and the reduced scattering region.
    for interface, extra in zip(interfaces, extras):
        extra_interface = set()
        if extra:
            extra_interface = set()
            for i, j in syst.graph:
                if i in extra and j not in extra:
                    extra_interface.add(j)
            interface -= extra
            interface |= extra_interface
        assert not extra.intersection(interface)

    pre_msg = (
        'Attaching leads results in gauge-fixed loops in the extended '
        'scattering region (scattering region plus one lead unit cell '
        'from every lead). This does not allow consistent gauge-fixing.\n\n'
    )
    solution_msg = (
        'To avoid this error, attach leads further away from each other.\n\n'
        'Note: calling `attach_lead()` with `add_cells > 0` will not fix '
        'this problem, as the added sites inherit the gauge from the lead. '
        'To extend the scattering region, you must manually add sites '
        'making sure that they use the scattering region gauge.'
    )

    # Check that there is at most one overlapping site between
    # reduced interface of one lead and extra sites of another
    num_leads = len(syst.leads)
    metagraph = scipy.sparse.lil_matrix((num_leads, num_leads))
    for i, j in permutations(range(num_leads), 2):
        intersection = len(interfaces[i] & (interfaces[j] | extras[j]))
        if intersection > 1:
            raise ValueError(
                pre_msg
                + ('There is at least one gauge-fixed loop in the overlap '
                   'of leads {} and {}.\n\n'.format(i, j))
                + solution_msg
            )
        elif intersection == 1:
            metagraph[i, j] = 1
    # Check that there is no loop formed by gauge-fixed bonds of multiple leads.
    num_components = scipy.sparse.csgraph.connected_components(metagraph, return_labels=False)
    if metagraph.nnz // 2 + num_components != num_leads:
        raise ValueError(
            pre_msg
            + ('There is at least one gauge-fixed loop formed by more than 2 leads. '
               ' The connectivity matrix of the leads is:\n\n'
               '{}\n\n'.format(metagraph.A))
            + solution_msg
        )

### Phase calculation

def _calculate_phases(loops, pos, previous_phase, flux):
    """Calculate the phase across the terminal links of a set of loops

    Parameters
    ----------
    loops : sequence of sequences of int
        The loops over which to calculate the flux. We wish to find the phase
        over the link connecting the first and last sites in each loop.
        The phase across all other links in a given loop is assumed known.
    pos : callable : int -> ndarray
        A map from site (integer) to realspace position.
    previous_phase : callable
        Takes a dict that maps from links to phases, and a loop and returns
        the product of the phases across each link in the loop,
        *except* the link between the first and last site in the loop.
    flux : callable
        Takes a sequence of positions and returns the magnetic flux through the
        surface defined by the provided loop.

    Returns
    -------
    phases : dict : (int, int) -> float
        A map from links to the phase across those links.
    """
    phases = dict()
    for loop in loops:
        tail, head = loop[-1], loop[0]
        integral = flux([pos(p) for p in loop])
        phase = np.exp(1j * np.pi * integral)
        phases[tail, head] = phase / previous_phase(phases, loop)
    return phases


# These functions are to be used with '_calculate_phases'.
# 'phases' always stores *either* the phase across (i, j) *or*
# (j, i), and never both. If a phase is not present it is assumed to
# be zero.


def _previous_phase_finite(phases, path):
    previous_phase = 1
    for i, j in zip(path, path[1:]):
        previous_phase *= phases.get((i, j), 1)
        previous_phase /= phases.get((j, i), 1)
    return previous_phase


def _previous_phase_infinite(cell_size, phases, path):
    previous_phase = 1
    for i, j in zip(path, path[1:]):
        # i and j are only in the fundamental unit cell (0 <= i < cell_size)
        # or the next one (cell_size <= i < 2 * cell_size).
        if i >= cell_size and j >= cell_size:
            assert i // cell_size == j // cell_size
            i = i % cell_size
            j = j % cell_size
        previous_phase *= phases.get((i, j), 1)
        previous_phase /= phases.get((j, i), 1)
    return previous_phase


def _previous_phase_composite(which_patch, extended_sites, lead_phases,
                              phases, path):
    previous_phase = 1
    for i, j in zip(path, path[1:]):
        patch_i = which_patch(i)
        patch_j = which_patch(j)
        if patch_i == -1 and patch_j == -1:
            # Both sites in reduced scattering region.
            previous_phase *= phases.get((i, j), 1)
            previous_phase /= phases.get((j, i), 1)
        else:
            # At least one site in a lead patch; use the phase from the
            # associated lead. Check that if both are in a patch, they
            # are in the same patch.
            assert patch_i * patch_j <= 0 or patch_i == patch_j
            patch = max(patch_i, patch_j)
            a, b = extended_sites(i), extended_sites(j)
            previous_phase *= lead_phases[patch](a, b)

    return previous_phase


### High-level interface
#
# These functions glue all the above functionality together.

# Wrapper for phase dict that takes high-level sites
def _finite_wrapper(syst, phases, a, b):
    i = syst.id_by_site[a]
    j = syst.id_by_site[b]
    # We only store *either* (i, j) *or* (j, i). If not present
    # then the phase is unity by definition.
    if (i, j) in phases:
        return phases[i, j]
    elif (j, i) in phases:
        return phases[j, i].conjugate()
    else:
        return 1


def _infinite_wrapper(syst, phases, a, b):
    sym = syst.symmetry
    # Bring link to fundamental domain consistently with how
    # we store the phases.
    t = max(sym.which(a), sym.which(b))
    a, b = sym.act(-t, a, b)
    i = syst.id_by_site[a]
    j = syst.id_by_site[b]
    # We only store *either* (i, j) *or* (j, i). If not present
    # then the phase is unity by definition.
    if (i, j) in phases:
        return phases[i, j]
    elif (j, i) in phases:
        return phases[j, i].conjugate()
    else:
        return 1


def _peierls_finite(syst, loops, syst_field, tol, average):
    integrate = partial(_surface_integral, syst_field,
                        tol=tol, average=average)
    phases = _calculate_phases(
        loops,
        syst.pos,
        _previous_phase_finite,
        integrate,
    )
    return partial(_finite_wrapper, syst, phases)


def _peierls_infinite(syst, loops, extended_sites, syst_field, tol, average):
    integrate = partial(_surface_integral, syst_field,
                        tol=tol, average=average)
    phases = _calculate_phases(
        loops,
        lambda i: extended_sites(i).pos,
        partial(_previous_phase_infinite, syst.cell_size),
        integrate,
    )
    return partial(_infinite_wrapper, syst, phases)


def _peierls_composite(syst, loops, which_patch, extended_sites, lead_gauges,
                       syst_field, *lead_fields, tol, average):
    if len(lead_fields) != len(syst.leads):
        raise ValueError('Magnetic fields must be provided for all leads.')

    lead_phases = [gauge(B, tol=tol, average=average)
                   for gauge, B in zip(lead_gauges, lead_fields)]

    flux = partial(_surface_integral, syst_field, tol=tol, average=average)

    # NOTE: uses the scattering region magnetic field to set the phase
    # of the inteface hoppings this choice is somewhat arbitrary,
    # but it is consistent with the position defined in the scattering
    # region coordinate system. the integrate functions for the leads
    # may be defined far from the interface.
    phases = _calculate_phases(
        loops,
        lambda i: extended_sites(i).pos,
        partial(_previous_phase_composite,
                which_patch, extended_sites, lead_phases),
        flux,
    )

    return (partial(_finite_wrapper, syst, phases), *lead_phases)


# This class is essentially a closure, but documenting closures is a pain.
# To emphasise the lack of manipulatable or inspectable state, we name the
# class as we would for a function.

class magnetic_gauge:
    """Fix the magnetic gauge for a finalized system.

    This can be used to later calculate the Peierls phases that
    should be applied to each hopping, given a magnetic field.

    This API is currently provisional. Refer to the documentation
    for details.

    Parameters
    ----------
    syst : `kwant.builder.FiniteSystem` or `kwant.builder.InfiniteSystem`

    Examples
    --------
    The following illustrates basic usage for a scattering region with
    a single lead attached:

    >>> import numpy as np
    >>> import kwant
    >>>
    >>> def hopping(a, b, t, peierls):
    >>>     return -t * peierls(a, b)
    >>>
    >>> syst = make_system(hopping)
    >>> lead = make_lead(hopping)
    >>> lead = lead.substituted(peierls='peierls_lead')
    >>> syst.attach_lead(lead)
    >>> syst = syst.finalized()
    >>>
    >>> gauge = kwant.physics.magnetic_gauge(syst)
    >>>
    >>> def B_syst(pos):
    >>>    return np.exp(-np.sum(pos * pos))
    >>>
    >>> peierls_syst, peierls_lead = gauge(B_syst, 0)
    >>>
    >>> params = dict(t=1, peierls=peierls_syst, peierls_lead=peierls_lead)
    >>> kwant.hamiltonian_submatrix(syst, params=params)
    """

    def __init__(self, syst):
        if isinstance(syst, builder.FiniteSystem):
            if syst.leads:
                loops, which_patch, extended_sites = _loops_in_composite(syst)
                lead_gauges = [magnetic_gauge(lead) for lead in syst.leads]
                self._peierls = partial(_peierls_composite, syst,
                                        loops, which_patch,
                                        extended_sites, lead_gauges)
            else:
                loops = _loops_in_finite(syst)
                self._peierls = partial(_peierls_finite, syst, loops)
        elif isinstance(syst, builder.InfiniteSystem):
            loops, extended_sites = _loops_in_infinite(syst)
            self._peierls = partial(_peierls_infinite, syst,
                                    loops, extended_sites)
        else:
            raise TypeError('Expected a finalized Builder')

    def __call__(self, syst_field, *lead_fields, tol=1E-8, average=False):
        """Return the Peierls phase for a magnetic field configuration.

        Parameters
        ----------
        syst_field : scalar, vector or callable
            The magnetic field to apply to the scattering region.
            If callable, takes a position and returns the
            magnetic field at that position. Can be a scalar if
            the system is 1D or 2D, otherwise must be a vector.
            Magnetic field is expressed in units :math:`φ₀ / l²`,
            where :math:`φ₀` is the magnetic flux quantum and
            :math:`l` is the unit of length.
        *lead_fields : scalar, vector or callable
            The magnetic fields to apply to each of the leads, in
            the same format as 'syst_field'. In addition, if a callable
            is provided, then the magnetic field must have the symmetry
            of the associated lead.
        tol : float, default: 1E-8
            The tolerance to which to calculate the flux through each
            hopping loop in the system.
        average : bool, default: False
            If True, estimate the magnetic flux through each hopping loop
            in the system by evaluating the magnetic field at a single
            position inside the loop and multiplying it by the area of the
            loop. If False, then ``scipy.integrate.quad`` is used to integrate
            the magnetic field. This parameter is only used when 'syst_field'
            or 'lead_fields' are callable.

        Returns
        -------
        phases : callable, or sequence of callables
            The first callable computes the Peierls phase in the scattering
            region and the remaining callables compute the Peierls phases
            in each of the leads. Each callable takes a pair of
            `~kwant.builder.Site` (a hopping) and returns a unit complex
            number (Peierls phase) that multiplies that hopping.
        """
        return self._peierls(syst_field, *lead_fields, tol=tol, average=False)



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

__all__ = ['TranslationalSymmetry', 'general', 'Polyatomic', 'Monatomic',
           'chain', 'square', 'cubic', 'triangular', 'honeycomb', 'kagome']

from math import sqrt
from itertools import product
import numpy as np
import tinyarray as ta
from . import builder
from .linalg import lll
from ._common import ensure_isinstance


def general(prim_vecs, basis=None, name='', norbs=None):
    """
    Create a Bravais lattice of any dimensionality, with any number of sites.

    Parameters
    ----------
    prim_vecs : 2d array-like of floats
        The primitive vectors of the Bravais lattice
    basis : 2d array-like of floats
        The coordinates of the basis sites inside the unit cell
    name : string or sequence of strings
        Name of the lattice, or sequence of names of all of the sublattices.
        If the name of the lattice is given, the names of sublattices (if any)
        are obtained by appending their number to the name of the lattice.
    norbs : int or sequence of ints, optional
        The number of orbitals per site on the lattice, or a sequence
        of the number of orbitals of sites on each of the sublattices.

    Returns
    -------
    lattice : either `Monatomic` or `Polyatomic`
        Resulting lattice.

    Notes
    -----
    This function is largely an alias to the constructors of corresponding
    lattices.
    """
    if basis is None:
        return Monatomic(prim_vecs, name=name, norbs=norbs)
    else:
        return Polyatomic(prim_vecs, basis, name=name, norbs=norbs)


def _check_prim_vecs(prim_vecs):
    """Check constraints to ensure that prim_vecs is correct."""
    if prim_vecs.ndim != 2:
        raise ValueError('``prim_vecs`` must be a 2d array-like object.')

    if prim_vecs.shape[0] > prim_vecs.shape[1]:
        raise ValueError('Number of primitive vectors exceeds '
                         'the space dimensionality.')

    if np.linalg.matrix_rank(prim_vecs) < len(prim_vecs):
        raise ValueError('"prim_vecs" must be linearly independent.')


class Polyatomic:
    """
    A Bravais lattice with an arbitrary number of sites in the basis.

    Contains `Monatomic` sublattices.  Note that an instance of ``Polyatomic`` is
    not itself a `~kwant.builder.SiteFamily`, only its sublattices are.

    Parameters
    ----------
    prim_vecs : 2d array-like of floats
        The primitive vectors of the Bravais lattice
    basis : 2d array-like of floats
        The coordinates of the basis sites inside the unit cell.
    name : string or sequence of strings, optional
        The name of the lattice, or a sequence of the names of all the
        sublattices.  If the name of the lattice is given, the names of
        sublattices are obtained by appending their number to the name of the
        lattice.
    norbs : int or sequence of ints, optional
        The number of orbitals per site on the lattice, or a sequence
        of the number of orbitals of sites on each of the sublattices.

    Raises
    ------
    ValueError
        If dimensionalities do not match.
    """
    def __init__(self, prim_vecs, basis, name='', norbs=None):
        prim_vecs = ta.array(prim_vecs, float)
        _check_prim_vecs(prim_vecs)

        dim = prim_vecs.shape[1]
        if name is None:
            name = ''
        if isinstance(name, str):
            name = [name + str(i) for i in range(len(basis))]

        basis = ta.array(basis, float)
        if basis.ndim != 2:
            raise ValueError('`basis` must be a 2d array-like object.')
        if basis.shape[1] != dim:
            raise ValueError('Basis dimensionality does not match '
                             'the space dimensionality.')

        try:
            norbs = list(norbs)
            if len(norbs) != len(basis):
                raise ValueError('Length of `norbs` is not the same as '
                                 'the number of basis vectors')
        except TypeError:
            norbs = [norbs] * len(basis)

        self.sublattices = [Monatomic(prim_vecs, offset, sname, norb)
                            for offset, sname, norb in zip(basis, name, norbs)]
        # Sequence of primitive vectors of the lattice.
        self._prim_vecs = prim_vecs
        # Precalculation of auxiliary arrays for real space calculations.
        self.reduced_vecs, self.transf = lll.lll(prim_vecs)
        self.voronoi = ta.dot(lll.voronoi(self.reduced_vecs), self.transf)

    def __str__(self):
        sl_names = ', '.join(str(sl.name) for sl in self.sublattices)
        return '<Polyatomic lattice with sublattices {0}>'.format(sl_names)

    def shape(self, function, start):
        """Return a key for all the lattice sites inside a given shape.

        The object returned by this method is primarily meant to be used as a
        key for indexing `~kwant.builder.Builder` instances.  See example below.

        Parameters
        ----------
        function : callable
            A function of real space coordinates that returns a truth value:
            true for coordinates inside the shape, and false otherwise.
        start : 1d array-like
            The real-space origin for the flood-fill algorithm.

        Returns
        -------
        shape_sites : function

        Notes
        -----
        When the function returned by this method is called, a flood-fill
        algorithm finds and yields all the lattice sites inside the specified
        shape starting from the specified position.

        A `~kwant.builder.Symmetry` or `~kwant.builder.Builder` may be passed as
        sole argument when calling the function returned by this method.  This
        will restrict the flood-fill to the fundamental domain of the symmetry
        (or the builder's symmetry).  Note that unless the shape function has
        that symmetry itself, the result may be unexpected.

        Examples
        --------
        >>> def circle(pos):
        ...     x, y = pos
        ...     return x**2 + y**2 < 100
        ...
        >>> lat = kwant.lattice.honeycomb()
        >>> syst = kwant.Builder()
        >>> syst[lat.shape(circle, (0, 0))] = 0
        >>> syst[lat.neighbors()] = 1
        """
        def shape_sites(symmetry=None):
            Site = builder.Site

            if symmetry is None:
                symmetry = builder.NoSymmetry()
            elif not isinstance(symmetry, builder.Symmetry):
                symmetry = symmetry.symmetry

            def fd_site(lat, tag):
                return symmetry.to_fd(Site(lat, tag, True))

            dim = len(start)
            if dim != self._prim_vecs.shape[1]:
                raise ValueError('Dimensionality of start position does not '
                                 'match the space dimensionality.')
            lats = self.sublattices
            deltas = list(self.voronoi)

            #### Flood-fill ####
            sites = []
            for tag in set(lat.closest(start) for lat in lats):
                for lat in lats:
                    site = fd_site(lat, tag)
                    if function(site.pos):
                        sites.append(site)
            if not sites:
                msg = 'No sites close to {0} are inside the desired shape.'
                raise ValueError(msg.format(start))
            tags = set(s.tag for s in sites)

            while sites:
                old_tags = tags
                tags = set()
                for site in sites:
                    yield site
                    tags.add(site.tag)
                old_tags |= tags

                new_tags = set()
                for tag in tags:
                    for delta in deltas:
                        new_tag = tag + delta
                        if new_tag not in old_tags:
                            new_tags.add(new_tag)

                sites = set()
                for tag in new_tags:
                    for lat in lats:
                        site = fd_site(lat, tag)
                        if site.tag not in old_tags and function(site.pos):
                            sites.add(site)

        return shape_sites

    def wire(self, center, radius):
        """Return a key for all the lattice sites inside an infinite cylinder.

        This method makes it easy to define cylindrical (2d: rectangular) leads
        that point in any direction.  The object returned by this method is
        primarily meant to be used as a key for indexing `~kwant.builder.Builder`
        instances.  See example below.

        Parameters
        ----------
        center : 1d array-like of floats
            A point belonging to the axis of the cylinder.
        radius : float
            The radius of the cylinder.

        Notes
        -----
        The function returned by this method is to be called with a
        `~kwant.builder.TranslationalSymmetry` instance (or a
        `~kwant.builder.Builder` instance whose symmetry is used then) as sole
        argument.  All the lattice sites (in the fundamental domain of the
        symmetry) inside the specified infinite cylinder are yielded.  The
        direction of the cylinder is determined by the symmetry.

        Examples
        --------
        >>> lat = kwant.lattice.honeycomb()
        >>> sym = kwant.TranslationalSymmetry(lat.a.vec((-2, 1)))
        >>> lead = kwant.Builder(sym)
        >>> lead[lat.wire((0, -5), 5)] = 0
        >>> lead[lat.neighbors()] = 1
        """
        center = ta.array(center, float)

        def wire_sites(sym):
            if not isinstance(sym, builder.Symmetry):
                sym = sym.symmetry
            if not isinstance(sym, TranslationalSymmetry):
                raise ValueError('wire shape only works with '
                                 'translational symmetry.')
            if len(sym.periods) != 1:
                raise ValueError('wire shape only works with one-dimensional '
                                 'translational symmetry.')
            period = np.array(sym.periods)[0]
            direction = ta.array(period / np.linalg.norm(period))
            r_squared = radius**2

            def wire_shape(pos):
                rel_pos = pos - center
                projection = rel_pos - direction * ta.dot(direction, rel_pos)
                return sum(projection * projection) <= r_squared

            return self.shape(wire_shape, center)(sym)

        return wire_sites

    def neighbors(self, n=1, eps=1e-8):
        """Return n-th nearest neighbor hoppings.

        Parameters
        ----------
        n : integer
            Order of the hoppings to return. Note that the zeroth neighbor is
            the site itself or any other sites with the same position.
        eps : float
            Tolerance relative to the length of the shortest lattice vector for
            when to consider lengths to be approximately equal.

        Returns
        -------
        hoppings : list of kwant.builder.HopplingKind objects
            The n-th nearest neighbor hoppings.

        Notes
        -----
        The hoppings are ordered lexicographically according to sublattice from
        which they originate, sublattice on which they end, and their lattice
        coordinates. Out of the two equivalent hoppings (a hopping and its
        reverse) only the lexicographically larger one is returned.
        """
        # This algorithm is not designed to be fast and can be improved,
        # however there is no real need.
        Site = builder.Site
        sls = self.sublattices
        shortest_hopping = sls[0].n_closest(
            sls[0].pos(([0] * sls[0].lattice_dim)), 2)[-1]
        rtol = eps
        eps *= np.linalg.norm(self.vec(shortest_hopping))
        nvec = len(self._prim_vecs)
        sublat_pairs = [(i, j) for (i, j) in product(sls, sls)
                        if sls.index(j) >= sls.index(i)]
        def first_nonnegative(tag):
            for i in tag:
                if i < 0:
                    return False
                elif i > 0:
                    return True
                else:
                    continue
            return True

        # Find the `n` closest neighbors (with multiplicity) for each
        # pair of lattices, this surely includes the `n` closest neighbors overall.
        sites = []
        for i, j in sublat_pairs:
            origin = Site(j, ta.zeros(nvec)).pos
            tags = i.n_closest(origin, n=n+1, group_by_length=True, rtol=rtol)
            ij_dist = [np.linalg.norm(Site(i, tag).pos - origin)
                          for tag in tags]
            sites.append((tags, (j, i), ij_dist))
        distances = np.r_[tuple((i[2] for i in sites))]
        distances = np.sort(distances)
        group_boundaries = np.where(np.diff(distances) > eps)[0]
        # Find distance in `n`-th group.
        if len(group_boundaries) == n:
            n_dist = distances[-1]
        else:
            n_dist = distances[group_boundaries[n]]

        # We now have all the required sites, select the ones in `n`-th group.
        result = []
        for group in sites:
            tags, distance = group[0], group[2]
            i, j = group[1]
            tags = np.array([tag for tag, dist in zip(tags, distance)
                             if abs(dist - n_dist) < eps])
            if len(tags):
                # Sort the tags.
                tags = tags[np.lexsort(tags.T[::-1])][::-1]
                # Throw away equivalent hoppings if
                # two sublattices are the same.
                if i == j and len(tags) > 1:
                    tags = tags[: len(tags) // 2]
                for tag in tags:
                    result.append(builder.HoppingKind(tag, j, i))
        return result

    @property
    def prim_vecs(self):
        """(sequence of vectors) Primitive vectors

        `prim_vecs[i]`` is the `i`-th primitive basis vector of the lattice
        displacement of the lattice origin from the real space coordinates
        origin.
        """
        return np.array(self._prim_vecs)

    def vec(self, int_vec):
        """
        Return the coordinates of a Bravais lattice vector in real space.

        Parameters
        ----------
        vec : integer vector

        Returns
        -------
        output : real vector
        """
        return ta.dot(int_vec, self._prim_vecs)


def short_array_repr(array):
    full = ' '.join([i.lstrip() for i in repr(array).split('\n')])
    return full[6 : -1]


def short_array_str(array):
    full = ', '.join([i.lstrip() for i in str(array).split('\n')])
    return full[1 : -1]


class Monatomic(builder.SiteFamily, Polyatomic):
    """
    A Bravais lattice with a single site in the basis.

    Instances of this class provide the `~kwant.builder.SiteFamily` interface.
    Site tags (see `~kwant.builder.SiteFamily`) are sequences of integers and
    describe the lattice coordinates of a site.

    ``Monatomic`` instances are used as site families on their own or as
    sublattices of `Polyatomic` lattices.

    Parameters
    ----------
    prim_vecs : 2d array-like of floats
        Primitive vectors of the Bravais lattice.

    offset : vector of floats
        Displacement of the lattice origin from the real space
        coordinates origin.

    Attributes
    ----------
    ``offset`` : vector
        Displacement of the lattice origin from the real space coordinates origin
    """

    def __init__(self, prim_vecs, offset=None, name='', norbs=None):
        prim_vecs = ta.array(prim_vecs, float)
        _check_prim_vecs(prim_vecs)

        dim = prim_vecs.shape[1]
        if name is None:
            name = ''

        if offset is None:
            offset = ta.zeros(dim)
        else:
            offset = ta.array(offset, float)
            if offset.shape != (dim,):
                raise ValueError('Dimensionality of offset does not match '
                                 'that of the space.')

        msg = '{0}({1}, {2}, {3}, {4})'
        cl = self.__module__ + '.' + self.__class__.__name__
        canonical_repr = msg.format(cl, short_array_repr(prim_vecs),
                                    short_array_repr(offset),
                                    repr(name), repr(norbs))
        super().__init__(canonical_repr, name, norbs)

        self.sublattices = [self]
        self._prim_vecs = prim_vecs
        self.inv_pv = ta.array(np.linalg.pinv(prim_vecs))
        self.offset = offset

        # Precalculation of auxiliary arrays for real space calculations.
        self.reduced_vecs, self.transf = lll.lll(prim_vecs)
        self.voronoi = ta.dot(lll.voronoi(self.reduced_vecs), self.transf)

        self.dim = dim
        self.lattice_dim = len(prim_vecs)

        if name != '':
            msg = "<Monatomic lattice {0}{1}>"
            orbs = ' with {0} orbitals'.format(self.norbs) if self.norbs else ''
            self.cached_str = msg.format(name, orbs)
        else:
            msg = "<unnamed Monatomic lattice, vectors {0}, origin [{1}]{2}>"
            orbs = ', with {0} orbitals'.format(norbs) if norbs else ''
            self.cached_str = msg.format(short_array_str(self._prim_vecs),
                                         short_array_str(self.offset), orbs)

    def __str__(self):
        return self.cached_str

    def normalize_tag(self, tag):
        tag = ta.array(tag, int)
        if len(tag) != self.lattice_dim:
            raise ValueError("Dimensionality mismatch.")
        return tag

    def n_closest(self, pos, n=1, group_by_length=False, rtol=1e-9):
        """Find n sites closest to position `pos`.

        Returns
        -------
        sites : numpy array
            An array with sites coordinates.
        """
        if len(pos) != len(self.offset):
            raise ValueError("Incorrect dimensionality of 'pos'.")

        # TODO (Anton): transform to tinyarrays, once ta indexing is better.
        return np.dot(lll.cvp(pos - self.offset, self.reduced_vecs,
                              n=n, group_by_length=group_by_length, rtol=rtol),
                      self.transf.T)

    def closest(self, pos):
        """
        Find the lattice coordinates of the site closest to position ``pos``.
        """
        return ta.array(self.n_closest(pos)[0])

    def pos(self, tag):
        """Return the real-space position of the site with a given tag."""
        return ta.dot(tag, self._prim_vecs) + self.offset


# The following class is designed such that it should avoid floating
# point precision issues.

class TranslationalSymmetry(builder.Symmetry):
    """A translational symmetry defined in real space.

    An alias exists for this common name: ``kwant.TranslationalSymmetry``.

    Group elements of this symmetry are integer tuples of appropriate length.

    Parameters
    ----------
    p0, p1, p2, ... : sequences of real numbers
        The symmetry periods in real space.

    Notes
    -----
    This symmetry automatically chooses the fundamental domain for each new
    `SiteFamily` it encounters. If this site family does not correspond to a
    Bravais lattice, or if it does not have a commensurate period, an error is
    produced. A certain flexibility in choice of the fundamental domain can be
    achieved by calling manually the `add_site_family` method and providing it
    the `other_vectors` parameter.

    The fundamental domain for hoppings are all hoppings ``(a, b)`` with site
    `a` in fundamental domain of sites.
    """
    def __init__(self, *periods):
        self.periods = ta.array(periods)
        if self.periods.ndim != 2:
            # TODO: remove the second part of the following message once
            # everybody got used to it.
            msg = ("TranslationalSymmetry takes 1d sequences as parameters.\n"
                   "See What's new in Kwant 0.2 in the documentation.")
            raise ValueError(msg)
        if np.linalg.matrix_rank(periods) < len(periods):
            raise ValueError("Translational symmetry periods must be "
                             "linearly independent")
        # A dictionary containing cached data required for applying the
        # symmetry to different site families.
        self.site_family_data = {}
        self.is_reversed = False

    def subgroup(self, *generators):
        """Return the subgroup generated by a sequence of group elements.

        Parameters
        ----------
        *generators: sequence of int
            Each generator must have length ``self.num_directions``.
        """
        generators = ta.array(generators)
        if generators.dtype != int:
            raise ValueError('Generators must be sequences of integers.')
        return TranslationalSymmetry(*ta.dot(generators, self.periods))

    def has_subgroup(self, other):
        if isinstance(other, builder.NoSymmetry):
            return True
        elif not isinstance(other, TranslationalSymmetry):
            raise ValueError("Unknown symmetry type.")

        if other.periods.shape[1] != self.periods.shape[1]:
            return False  # Mismatch of spatial dimensionalities.

        inv = np.linalg.pinv(self.periods)
        factors = np.dot(other.periods, inv)
        # Absolute tolerance is correct in the following since we want an error
        # relative to the closest integer.
        return (np.allclose(factors, np.round(factors), rtol=0, atol=1e-8) and
                np.allclose(ta.dot(factors, self.periods), other.periods))

    def add_site_family(self, fam, other_vectors=None):
        """
        Select a fundamental domain for site family and cache associated data.

        Parameters
        ----------
        fam : `SiteFamily`
            the site family which has to be processed.  Be sure to delete the
            previously processed site families from `site_family_data` if you
            want to modify the cache.

        other_vectors : 2d array-like of integers
            Bravais lattice vectors used to complement the periods in forming
            a basis. The fundamental domain consists of all the lattice sites
            for which the zero coefficients corresponding to the symmetry
            periods in the basis formed by the symmetry periods and
            `other_vectors`. If an insufficient number of `other_vectors` is
            provided to form a basis, the missing ones are selected
            automatically.

        Raises
        ------
        KeyError
            If `fam` is already stored in `site_family_data`.
        ValueError
            If lattice `fam` is incompatible with given periods.
        """
        ensure_isinstance(fam, Monatomic)

        dim = self.periods.shape[1]
        if fam in self.site_family_data:
            raise KeyError('Family already processed, delete it from '
                           'site_family_data first.')
        inv = np.linalg.pinv(fam.prim_vecs)
        try:
            bravais_periods = np.dot(self.periods, inv)
        except ValueError:
            fam_space_dim = fam.prim_vecs.shape[1]
            if dim == fam_space_dim:
                raise
            msg = ("{0}-d-embedded lattice is incompatible with "
                   "{1}-d translational symmetry.")
            raise ValueError(msg.format(fam_space_dim, dim))
        # Absolute tolerance is correct in the following since we want an error
        # relative to the closest integer.
        if (not np.allclose(bravais_periods, np.round(bravais_periods),
                            rtol=0, atol=1e-8) or
            not np.allclose([fam.vec(i) for i in bravais_periods],
                            self.periods)):
            msg = ('Site family {0} does not have commensurate periods with '
                   'symmetry {1}.')
            raise ValueError(msg.format(fam, self))
        bravais_periods = np.array(np.round(bravais_periods), dtype='int')
        (num_dir, lat_dim) = bravais_periods.shape
        if other_vectors is None:
            other_vectors = np.zeros((0, lat_dim), dtype=int)
        else:
            other_vectors = np.array(other_vectors)
            if other_vectors.ndim != 2:
                raise ValueError(
                    '`other_vectors` must be a 2d array-like object.')
            if np.any(np.round(other_vectors) - other_vectors):
                raise ValueError('Only integer other_vectors are allowed.')
            other_vectors = np.array(np.round(other_vectors), dtype=int)

        m = np.zeros((lat_dim, lat_dim), dtype=int)

        m.T[:num_dir] = bravais_periods
        num_vec = num_dir + len(other_vectors)
        m.T[num_dir:num_vec] = other_vectors

        if np.linalg.matrix_rank(m) < num_vec:
            raise ValueError('other_vectors and symmetry periods are not '
                             'linearly independent.')

        # To define the fundamental domain of the new site family we now need to
        # choose `lat_dim - num_vec` extra lattice vectors that are not
        # linearly dependent on the vectors we already have. To do so we
        # continuously add the lattice basis vectors one by one such that they
        # are not linearly dependent on the existent vectors
        while num_vec < lat_dim:
            vh = np.linalg.svd(np.dot(m[:, :num_vec].T, fam.prim_vecs),
                               full_matrices=False)[2]
            projector = np.identity(dim) - np.dot(vh.T, vh)

            residuals = np.dot(fam.prim_vecs, projector)
            residuals = np.apply_along_axis(np.linalg.norm, 1, residuals)
            m[np.argmax(residuals), num_vec] = 1
            num_vec += 1

        det_m = int(round(np.linalg.det(m)))
        if det_m == 0:
            raise RuntimeError('Adding site family failed.')

        det_x_inv_m = np.array(np.round(det_m * np.linalg.inv(m)), dtype=int)
        assert (np.dot(m, det_x_inv_m) // det_m == np.identity(lat_dim)).all()

        det_x_inv_m_part = det_x_inv_m[:num_dir, :]
        m_part = m[:, :num_dir]
        self.site_family_data[fam] = (ta.array(m_part),
                                     ta.array(det_x_inv_m_part), det_m)

    @property
    def num_directions(self):
        return len(self.periods)

    def _get_site_family_data(self, family):
        try:
            return self.site_family_data[family]
        except KeyError:
            self.add_site_family(family)
            return self.site_family_data[family]

    def which(self, site):
        det_x_inv_m_part, det_m = self._get_site_family_data(site.family)[-2:]
        result = ta.dot(det_x_inv_m_part, site.tag) // det_m
        return -result if self.is_reversed else result

    def act(self, element, a, b=None):
        element = ta.array(element)
        if element.dtype is not int:
            raise ValueError("group element must be a tuple of integers")
        m_part = self._get_site_family_data(a.family)[0]
        try:
            delta = ta.dot(m_part, element)
        except ValueError:
            msg = 'Expecting a {0}-tuple group element, but got `{1}` instead.'
            raise ValueError(msg.format(self.num_directions, element))
        if self.is_reversed:
            delta = -delta
        if b is None:
            return builder.Site(a.family, a.tag + delta, True)
        elif b.family == a.family:
            return (builder.Site(a.family, a.tag + delta, True),
                    builder.Site(b.family, b.tag + delta, True))
        else:
            m_part = self._get_site_family_data(b.family)[0]
            try:
                delta2 = ta.dot(m_part, element)
            except ValueError:
                msg = ('Expecting a {0}-tuple group element, '
                       'but got `{1}` instead.')
                raise ValueError(msg.format(self.num_directions, element))
            if self.is_reversed:
                delta2 = -delta2
            return (builder.Site(a.family, a.tag + delta, True),
                    builder.Site(b.family, b.tag + delta2, True))

    def reversed(self):
        """Return a reversed copy of the symmetry.

        The resulting symmetry has all the period vectors opposite to the
        original and an identical fundamental domain.
        """
        result = TranslationalSymmetry(*self.periods)
        result.site_family_data = self.site_family_data
        result.is_reversed = not self.is_reversed
        result.periods = -self.periods
        return result



################ Library of lattices

def chain(a=1, name='', norbs=None):
    """Make a one-dimensional lattice."""
    return Monatomic(((a,),), name=name, norbs=norbs)


def square(a=1, name='', norbs=None):
    """Make a square lattice."""
    return Monatomic(((a, 0), (0, a)), name=name, norbs=norbs)


def cubic(a=1, name='', norbs=None):
    """Make a cubic lattice."""
    return Monatomic(((a, 0, 0), (0, a, 0), (0, 0, a)),
                     name=name, norbs=norbs)


tri = ta.array(((1, 0), (0.5, 0.5 * sqrt(3))))

def triangular(a=1, name='', norbs=None):
    """Make a triangular lattice."""
    return Monatomic(a * tri, name=name, norbs=norbs)


def honeycomb(a=1, name='', norbs=None):
    """Make a honeycomb lattice."""
    lat = Polyatomic(a * tri, ((0, 0), (0, a / sqrt(3))),
                     name=name, norbs=norbs)
    lat.a, lat.b = lat.sublattices
    return lat


def kagome(a=1, name='', norbs=None):
    """Make a kagome lattice."""
    lat = Polyatomic(a * tri, ((0, 0),) + tuple(0.5 * a * tri),
                     name=name, norbs=norbs)
    lat.a, lat.b, lat.c = lat.sublattices
    return lat


# TODO (Anton): unhide _prim_vecs, once tinyarray supports indexing of
# sub-arrays.
